#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
plasma_calc.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr
from astropy import constants

from .resample import resample


def plasma_calc(b_xyz=None, t_i=None, t_e=None, n_i=None, n_e=None):
    """Computes plasma parameters including characteristic length and time scales.
    
    Parameters
    ----------
    b_xyz : xarray.DataArray
        Time series of the magnetic field [nT].
    
    t_i : xarray.DataArray
        Time series of the ions scalar temperature [eV].
    
    t_e : xarray.DataArray
        Time series of the electrons scalar temperature [eV].
    
    n_i : xarray.DataArray
        Time series of the ions number density [cm^{-3}].
    
    n_e : xarray.DataArray
        Time series of the electrons number density [cm^{-3}].

    Returns
    -------
    out : xarray.Dataset
        Dataset of the plasma parameters :
            * time : xarray.DataArray
                Time.

            * Wpe : xarray.DataArray
                Time series of the electron plasma frequency [rad.s^{-1}].

            * Fpe : xarray.DataArray
                Time series of the electron plasma frequency [Hz].

            * Wce : xarray.DataArray
                Time series of the electron cyclotron frequency [rad.s^{-1}].

            * Fce : xarray.DataArray
                Time series of the electron cyclotron frequency [Hz].

            * Wpp : xarray.DataArray
                Time series of the ion plasma frequency [rad.s^{-1}].

            * Fpp : xarray.DataArray
                Time series of the ion plasma frequency [Hz].

            * Fcp : xarray.DataArray
                Time series of the ion cyclotron frequency [Hz].

            * Fuh : xarray.DataArray
                Time series of the upper hybrid frequency [Hz].

            * Flh : xarray.DataArray
                Time series of the lower hybrid frequency [Hz].

            * Va : xarray.DataArray
                Time series of the Alfvèn velocity (ions) [m.s^{-1}].

            * Vae : xarray.DataArray
                Time series of the Alfvèn velocity (electrons) [m.s^{-1}].

            * Vte : xarray.DataArray
                Time series of the electron thermal velocity [m.s^{-1}].

            * Vtp : xarray.DataArray
                Time series of the electron thermal velocity [m.s^{-1}].

            * Vts : xarray.DataArray
                Time series of the sound speed [m.s^{-1}].

            * gamma_e : xarray.DataArray
                Time series of the electron Lorentz factor.

            * gamma_p : xarray.DataArray
                Time series of the electron Lorentz factor.

            * Le : xarray.DataArray
                Time series of the electron inertial length [m].

            * Li : xarray.DataArray
                Time series of the electron inertial length [m].

            * Ld : xarray.DataArray
                Time series of the Debye length [m].

            * Nd : xarray.DataArray
                Time series of the number of electrons in the Debye sphere.

            * Roe : xarray.DataArray
                Time series of the electron Larmor radius [m].

            * Rop : xarray.DataArray
                Time series of the ion Larmor radius [m].

            * Ros : xarray.DataArray
                Time series of the length associated to the sound speed [m].

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field, ion/electron temperature and number density
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> t_xyz_i = mms.get_data("Ti_gse_fpi_fast_l2", tint, mms_id)
    >>> t_xyz_e = mms.get_data("Te_gse_fpi_fast_l2", tint, mms_id)
    >>> n_i = mms.get_data("Ni_fpi_fast_l2", tint, mms_id)
    >>> n_e = mms.get_data("Ne_fpi_fast_l2", tint, mms_id)
    >>> # Compute scalar temperature
    >>> t_xyzfac_i = mms.rotate_tensor(t_xyz_i, "fac", b_xyz, "pp")
    >>> t_xyzfac_e = mms.rotate_tensor(t_xyz_e, "fac", b_xyz, "pp")
    >>> t_i = pyrf.trace(t_xyzfac_i)
    >>> t_e = pyrf.trace(t_xyzfac_e)
    >>> # Compute plasma parameters
    >>> plasma_params = pyrf.plasma_calc(b_xyz, t_i, t_e, n_i, n_e)

    """

    assert b_xyz is not None and isinstance(b_xyz, xr.DataArray)
    assert t_i is not None and isinstance(t_i, xr.DataArray)
    assert t_e is not None and isinstance(t_e, xr.DataArray)
    assert n_i is not None and isinstance(n_i, xr.DataArray)
    assert n_e is not None and isinstance(n_e, xr.DataArray)

    # Get constants
    e, c = [constants.e.value, constants.c.value]
    mu0, eps0 = [constants.mu0.value, constants.eps0.value]
    m_p, m_e = [constants.m_p.value, constants.m_e.value]
    mp_me = m_p / m_e

    # Resample all variables with respect to the magnetic field
    n_t = len(b_xyz)

    if len(t_i) != n_t:
        t_i = resample(t_i, b_xyz).data

    if len(t_e) != n_t:
        t_e = resample(t_e, b_xyz).data

    if len(n_i) != n_t:
        n_i = resample(n_i, b_xyz).data

    if len(n_e) != n_t:
        n_e = resample(n_e, b_xyz).data

    # Transform number density and magnetic field to SI units
    n_i, n_e = [1e6 * n_i, 1e6 * n_e]

    if b_xyz.ndim == 2:
        b_si = 1e-9 * np.linalg.norm(b_xyz, axis=1)
    else:
        b_si = 1e-9 * np.linalg.norm(b_xyz, axis=1)

    w_pe = np.sqrt(n_e * e ** 2 / (m_e * eps0)) 	# rad/s
    w_ce = e * b_si / m_e   						# rad/s
    w_pp = np.sqrt(n_i * e ** 2 / (m_p * eps0))

    v_a = b_si / np.sqrt(mu0 * n_i * m_p)

    v_ae = b_si / np.sqrt(mu0 * n_e * m_e)
    v_te = c * np.sqrt(1 - 1 / (t_e * e / (m_e * c ** 2) + 1) ** 2) 	# m/s (relativ. correct)
    v_tp = c * np.sqrt(1 - 1 / (t_i * e / (m_p * c ** 2) + 1) ** 2)     # m/s
    # Sound speed formula (F. Chen, Springer 1984).
    v_ts = np.sqrt((t_e * e + 3 * t_i * e) / m_p)

    gamma_e = 1 / np.sqrt(1 - (v_te / c) ** 2)
    gamma_p = 1 / np.sqrt(1 - (v_tp / c) ** 2)

    l_e = c / w_pe
    l_i = c / w_pp
    # Debye length scale, sqrt(2) needed because of Vte definition
    l_d = v_te / (w_pe * np.sqrt(2))
    # number of e- in Debye sphere
    n_d = l_d * eps0 * m_e * v_te ** 2 / e ** 2

    f_pe = w_pe / (2 * np.pi) 				# Hz
    f_ce = w_ce / (2 * np.pi)
    f_uh = np.sqrt(f_ce ** 2 + f_pe ** 2)
    f_pp = w_pp / (2 * np.pi)
    f_cp = f_ce / mp_me
    f_lh = np.sqrt(f_cp * f_ce / (1 + f_ce ** 2 / f_pe ** 2) + f_cp ** 2)

    rho_e = m_e * c / (e * b_si) * np.sqrt(gamma_e ** 2 - 1) 	# m, relativistically correct
    rho_p = m_p * c / (e * b_si) * np.sqrt(gamma_p ** 2 - 1) 	# m, relativistically correct
    rho_s = v_ts / (f_cp * 2 * np.pi) 							# m

    out = xr.Dataset(
        {"time": b_xyz.time.data, "w_pe": (["time"], w_pe), "w_ce": (["time"], w_ce),
         "w_pp": (["time"], w_pp),
         "v_a": (["time"], v_a), "v_ae": (["time"], v_ae), "v_te": (["time"], v_te),
         "v_tp": (["time"], v_tp), "v_ts": (["time"], v_ts), "gamma_e": (["time"], gamma_e),
         "gamma_p": (["time"], gamma_p), "l_e": (["time"], l_e), "l_i": (["time"], l_i),
         "l_d": (["time"], l_d), "n_d": (["time"], n_d), "f_pe": (["time"], f_pe),
         "f_ce": (["time"], f_ce), "f_uh": (["time"], f_uh), "f_pp": (["time"], f_pp),
         "f_cp": (["time"], f_cp), "f_lh": (["time"], f_lh), "rho_e": (["time"], rho_e),
         "rho_p": (["time"], rho_p), "rho_s": (["time"], rho_s)})

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_eis_omni.py

@author : Louis RICHARD
"""

import numpy as np

from .list_files import list_files
from .db_get_ts import db_get_ts


def get_eis_omni(tar_var="Flux_extof_proton_srvy_l2", tint=None, mms_id=2, verbose=True):
    """Computes omni directional energy spectrum of the target data unit for the target specie
    over the target energy range.

    Parameters
    ----------
    tar_var : str
        Key of the target variable like {data_unit}_{dtype}_{specie}_{data_rate}_{data_lvl}.

    tint : list of str
        Time interval.

    mms_id : int or float or str
        Index of the spacecraft.

    verbose : bool
        Set to True to follow the loading. Default is True.

    Returns
    --------
    out : xarray.DataArray
        Energy spectrum of the target data unit for the target specie in omni direction.

    """

    assert isinstance(tar_var, str)
    assert tint is not None and isinstance(tint, list)
    assert isinstance(tint[0], str) and isinstance(tint[1], str)
    assert isinstance(mms_id, (int, str)) and int(mms_id) in np.arange(1, 5)
    assert isinstance(verbose, bool)

    mms_id = int(mms_id)

    data_unit, data_type, specie, data_rate, data_lvl = tar_var.split("_")

    var = {"mms_id": mms_id, "inst": "epd-eis"}

    pref = f"mms{mms_id:d}_epd_eis"

    if data_rate == "brst":
        var["tmmode"] = data_rate
        pref = f"{pref}_{data_rate}"

    elif data_rate == "srvy":
        var["tmmode"] = data_rate

    else:
        raise ValueError("Invalid data rate")

    var["lev"] = data_lvl

    if data_type == "electronenergy":
        if specie == "electron":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"{pref}_{data_type}_{specie}"
        else:
            raise ValueError("invalid specie")
    elif data_type == "extof":
        if specie == "proton":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"{pref}_{data_type}_{specie}"
        elif specie == "oxygen":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"{pref}_{data_type}_{specie}"
        elif specie == "alpha":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"{pref}_{data_type}_{specie}"
        else:
            raise ValueError("invalid specie")
    elif data_type == "phxtof":
        if specie == "proton":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"{pref}_{data_type}_{specie}"
        elif specie == "oxygen":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"{pref}_{data_type}_{specie}"
        else:
            raise ValueError("Invalid specie")
    else:
        raise ValueError("Invalid data type")

    files = list_files(tint, mms_id, var)

    file_version = int(files[0].split("_")[-1][1])

    var["version"] = file_version

    if data_unit.lower() in ["flux", "counts", "cps"]:
        suf = f"P{file_version:d}_{data_unit.lower()}_t"
    else:
        raise ValueError("Invalid data unit")

    dset_name = f"mms{var['mms_id']:d}_{var['inst']}_{var['tmmode']}_{var['lev']}_{var['dtype']}"
    cdf_names = [f"{pref}_{suf}{t:d}" for t in range(6)]

    out_dict = {}

    flux_omni = None

    for i, cdfname in enumerate(cdf_names):
        scope_key = f"t{i:d}"

        if verbose:
            print(f"Loading {cdfname}...")

        out_dict[scope_key] = db_get_ts(dset_name, cdfname, tint)
        try:
            flux_omni += out_dict[scope_key]
        except TypeError:
            flux_omni = out_dict[scope_key]

    flux_omni /= 6

    return flux_omni



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
feeps_pitch_angles.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy.time import Time

from .get_feeps_active_eyes import get_feeps_active_eyes


def feeps_pitch_angles(inp_dataset=None, b_bcs=None):
    """Computes the FEEPS pitch angles for each telescope from magnetic field data.

    Parameters
    ----------
    inp_dataset : xarray.Dataset
        Dataset of the time series of the energy spectrum for each eye of FEEPS telescopes.

    b_bcs : xarray.DataArray
        Time series of the magnetic in spacecraft coordinates system.

    Returns
    -------
    out : xarray.DataArray
        Time series of the pitch angles.

    idx_maps : dict
        to fill.

    """

    assert inp_dataset is not None and isinstance(inp_dataset, xr.Dataset)
    assert b_bcs is not None and isinstance(b_bcs, xr.DataArray)

    var = inp_dataset.attrs
    # mms_id = var["mmsId"]
    times = inp_dataset.time
    b_times = b_bcs.time

    tint = Time(np.hstack([times.data.min(), times.data.max()]), format="datetime64").isot

    eyes = get_feeps_active_eyes(var, tint, var["mmsId"])

    idx_maps = None
    bins = 13  # number of pitch angle bins; 10 deg = 17 bins, 15 deg = 13 bins
    # dpa = 180.0 / bins  # delta-pitch angle for each bin

    # Rotation matrices for FEEPS coord system (FCS) into body coordinate system (BCS):
    t_top = [[1./np.sqrt(2.), -1./np.sqrt(2.), 0], [1./np.sqrt(2.), 1./np.sqrt(2.), 0], [0, 0, 1]]
    t_bot = [[-1. / np.sqrt(2.), -1. / np.sqrt(2.), 0], [-1. / np.sqrt(2.), 1. / np.sqrt(2.), 0],
             [0, 0, -1]]

    # Telescope vectors in FCS:
    v_fcs = {"1": [0.347, -0.837, 0.423], "2": [0.347, -0.837, -0.423], "3": [0.837, -0.347, 0.423],
             "4": [0.837, -0.347, -0.423], "5": [-0.087, 0.000, 0.996], "6": [0.104, 0.180, 0.978],
             "7": [0.654, -0.377, 0.656], "8": [0.654, -0.377, -0.656], "9": [0.837, 0.347, 0.423],
             "10": [0.837, 0.347, -0.423], "11": [0.347, 0.837, 0.423],
             "12": [0.347, 0.837, -0.423]}

    telescope_map = {"bottom-electron": [1, 2, 3, 4, 5, 9, 10, 11, 12], "bottom-ion": [6, 7, 8],
                     "top-electron": [1, 2, 3, 4, 5, 9, 10, 11, 12], "top-ion": [6, 7, 8]}

    top_tele_idx_map, bot_tele_idx_map = [{}, {}]

    pas = np.empty([len(b_times), 18])  # pitch angles for each eye at each time

    # Telescope vectors in Body Coordinate System:
    # Factors of -1 account for 180 deg shift between particle velocity and telescope normal
    # direction :
    # Top:
    vt_bcs, vb_bcs = [{}, {}]

    for s in telescope_map["top-{}".format(var["dtype"])]:
        s = str(s)

        vt_bcs[s] = [
                     -1. * (t_top[0][0] * v_fcs[s][0] + t_top[0][1] * v_fcs[s][1] + t_top[0][2] *
                            v_fcs[s][2]),
                     -1. * (t_top[1][0] * v_fcs[s][0] + t_top[1][1] * v_fcs[s][1] + t_top[1][2] *
                            v_fcs[s][2]),
                     -1. * (t_top[2][0] * v_fcs[s][0] + t_top[2][1] * v_fcs[s][1] + t_top[2][2] *
                            v_fcs[s][2])]

    for s in telescope_map["bottom-{}".format(var["dtype"])]:
        s = str(s)

        vb_bcs[s] = [
                     -1. * (t_bot[0][0] * v_fcs[s][0] + t_bot[0][1] * v_fcs[s][1] + t_bot[0][2] *
                            v_fcs[s][2]),
                     -1. * (t_bot[1][0] * v_fcs[s][0] + t_bot[1][1] * v_fcs[s][1] + t_bot[1][2] *
                            v_fcs[s][2]),
                     -1. * (t_bot[2][0] * v_fcs[s][0] + t_bot[2][1] * v_fcs[s][1] + t_bot[2][2] *
                            v_fcs[s][2])]

    telescopes = np.hstack([telescope_map["bottom-{}".format(var["dtype"])],
                            telescope_map["top-{}".format(var["dtype"])]])

    for i, k in zip(np.arange(18), telescopes):
        if i < 8:
            v_bcs = vt_bcs[str(k)]
        else:
            v_bcs = vb_bcs[str(k)]

        # pas[:, i] = 180. / math.pi * np.arccos(
        # 	(v_bcs[0] * b_bcs[:, 0] + v_bcs[1] * b_bcs[:, 1] + v_bcs[2] * b_bcs[:, 2]) / (
        # 				np.sqrt(v_bcs[0] ** 2 + v_bcs[1] ** 2 + v_bcs[2] ** 2) * np.sqrt(
        # 			b_bcs[:, 0] ** 2 + b_bcs[:, 1] ** 2 + b_bcs[:, 2] ** 2)))

        if var["tmmode"] == "srvy":
            if i < 8:
                top_tele_idx_map[k] = i
            else:
                bot_tele_idx_map[k] = i

            top_idx, bot_idx = [[], []]

            # PAs for only active eyes
            # pitch angles for each eye at each time
            new_pas = np.empty([len(b_times), len(eyes["top"]) + len(eyes["bottom"])])

            for top_idx, top_eye in enumerate(eyes["top"]):
                new_pas[:, top_idx] = pas[:, top_tele_idx_map[top_eye]]
                top_idx.append(top_idx)

            for bot_idx, bot_eye in enumerate(eyes["bottom"]):
                new_pas[:, bot_idx + len(eyes["top"])] = pas[:, bot_tele_idx_map[bot_eye]]
                bot_idx.append(bot_idx + len(eyes["top"]))

            idx_maps = {"electron-top": top_idx, "electron-bottom": bot_idx}

        else:
            new_pas = pas

    out_data = xr.DataArray(new_pas, coords=[b_times, np.arange(18)], dims=["time", "idx"])

    # interpolate to the PA time stamps
    out = out_data.interp({'time': times})

    return out, idx_maps



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
movmean.py

@author : Louis RICHARD
"""

import xarray as xr
import numpy as np


def movmean(inp=None, n_pts=100):
    """Computes running average of the inp over npts points.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input variable.

    n_pts : int
        Number of points to average over.

    Returns
    -------
    out : xarray.DataArray
        Time series of the input variable averaged over npts points.

    Notes
    -----
    Works also with 3D skymap distribution.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000","2019-09-14T08:11:00.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load ion pressure tensor
    >>> p_xyz_i = mms.get_data("Pi_gse_fpi_brst_l2", tint, mms_id)
    >>> # Running average the pressure tensor over 10s
    >>> fs = pyrf.calc_fs(p_xyz_i)
    >>>> p_xyz_i = pyrf.movmean(p_xyz_i, int(10 * fs))

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert isinstance(n_pts, (float, int))

    if isinstance(n_pts, float):
        n_pts = np.floor(n_pts).astype(int)

    if n_pts % 2:
        n_pts -= 1

    # Computes moving average
    cum_sum = np.cumsum(inp.data, axis=0)
    out_dat = (cum_sum[n_pts:, ...] - cum_sum[:-n_pts, ...]) / n_pts

    coords = []

    for k in inp.dims:
        if k == "time":
            coords.append(inp.coords[k][int(n_pts / 2):-int(n_pts / 2)])
        else:
            coords.append(inp.coords[k].data)

    # Output in DataArray type
    out = xr.DataArray(out_dat, coords=coords, dims=inp.dims, attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
wavepolarize_means.py

@author : Louis RICHARD
"""

import warnings
import numpy as np
import xarray as xr
from astropy.time import Time

from .resample import resample


def wavepolarize_means(b_wave=None, b_bgd=None, **kwargs):
    """
    Analysis the polarization of magnetic wave using "means" method

    Parameters
    -----------
    b_wave : xarray.DataArray
        Time series of the magnetic field from Search Coil Magnetometer (SCM).

    b_bgd : xarray.DataArray
        Time series of the magnetic field from Flux Gate Magnetometer (FGM).

    **kwargs : dict
        Hash table of keyword arguments with :
            * min_psd : float
                Threshold for the analysis (e.g 1.0e-7). Below this value, the SVD analysis is
                meaningless if min_psd is not given, SVD analysis will be done for all waves.
                Default ``min_psd`` = 1e-25.

            * nop_fft : int
                Number of points in FFT. (default 256).

    Returns
    -------
    b_psd : xarray.DataArray
        Power spectrum density of magnetic filed wave.

    wave_angle : xarray.DataArray
        (form 0 to 90)

    deg_pol : xarray.DataArray
        Spectrogram of the degree of polarization (form 0 to 1).

    elliptict : xarray.DataArray
        Spectrogram of the ellipticity (form -1 to 1)

    helict : xarray.DataArray
        Spectrogram of the helicity (form -1 to 1)

    Notes
    ------
    ``b_wave`` and ``b_bgd`` should be from the same satellite and in the same coordinates

    .. warning::
        If one component is an order of magnitude or more  greater than the other two then the
        polarization results saturate and erroneously indicate high degrees of polarization at
        all times and frequencies. Time series should be eyeballed before running the program.
        For time series containing very rapid changes or spikes the usual problems with Fourier
        analysis arise. Care should be taken in evaluating degree of polarization results. For
        meaningful results there should be significant wave power at the frequency where the
        polarization approaches 100%. Remember comparing two straight lines yields 100%
        polarization.

    Examples
    --------
    >>> from pyrfu import pyrf
    >>> polarization = pyrf.wavepolarize_means(b_wave, b_bgd)
    >>> polarization = pyrf.wavepolarize_means(b_wave, b_bgd, 1.0e-7)
    >>> polarization = pyrf.wavepolarize_means(b_wave, b_bgd, 1.0e-7, 256)

    """

    if (b_wave is None) or (b_bgd is None):
        raise ValueError("wavepolarize_means requires at least two arguments")

    min_psd = 1e-25
    nop_fft = 256

    if kwargs.get("min_psd"):
        min_psd = kwargs["min_psd"]

    if kwargs.get("nop_fft"):
        nop_fft = kwargs["nop_fft"]

    step_length = int(nop_fft / 2)
    n_pts = len(b_wave)
    n_stp = int((n_pts - nop_fft) / step_length)  				# total number of FFTs
    n_bin = 7  													# No. of bins in frequency domain
    aa = np.array([.024, .093, .232, .301, .232, .093, .024])  	# Smoothing profile based on Hanning

    # change wave to MFA coordinates
    b_bgd = resample(b_bgd, b_wave)

    b_x, b_y, b_z = [np.zeros(len(b_wave)) for _ in range(3)]

    for ii in range(len(b_wave)):
        nb = b_bgd[ii, :] / np.linalg.norm(b_bgd[ii, :])
        n_perp1 = np.cross(nb, [0, 1, 0])
        n_perp1 = n_perp1 / np.linalg.norm(n_perp1)
        n_perp2 = np.cross(nb, n_perp1)

        b_z[ii] = np.sum(b_wave[ii, :] * nb)
        b_x[ii] = np.sum(b_wave[ii, :] * n_perp1)
        b_y[ii] = np.sum(b_wave[ii, :] * n_perp2)

    ct = Time(b_wave.time.data, format="datetime64").unix

    # DEFINE ARRAYS
    xs, ys, zs = [b_x, b_y, b_z]

    sample_freq = 1/(ct[1]-ct[0])
    end_sample_freq = 1/(ct[-1]-ct[-2])

    if sample_freq != end_sample_freq:
        warnings.warn(
            "file sampling frequency changes {} Hz to {} Hz".format(sample_freq, end_sample_freq),
            UserWarning)
    else:
        print("ac file sampling frequency {} Hz".format(sample_freq))

    # FFT calculation
    # Minimum variance direction and wave normal angle
    wave_angle = np.zeros([n_stp, int(nop_fft / 2)])

    # Degree of Polarization
    sqrd_mat = np.zeros([n_stp, int(nop_fft / 2), 3, 3])
    trace_sqrd_mat, trace_spec_mat, deg_pol = [np.zeros([n_stp, int(nop_fft / 2)]) for _ in
                                               range(3)]

    # HELICITY, ELLIPTICITY AND THE WAVE STATE VECTOR
    lambda_u = np.zeros([n_stp, int(nop_fft / 2), 3, 3])

    helic, ellip = [np.zeros([n_stp, int(nop_fft / 2), 3]) for _ in range(2)]

    for j in range(n_stp):
        # FFT CALCULATION
        smooth = 0.08 + 0.46 * (1 - np.cos(2 * np.pi * np.arange(1, nop_fft + 1) / nop_fft))
        temp_x = smooth * xs[((j - 1) * step_length + 1):((j - 1) * step_length + nop_fft)]
        temp_y = smooth * ys[((j - 1) * step_length + 1):((j - 1) * step_length + nop_fft)]
        temp_z = smooth * zs[((j - 1) * step_length + 1):((j - 1) * step_length + nop_fft)]

        spec_x = np.fft.fft(temp_x)
        spec_y = np.fft.fft(temp_y)
        spec_z = np.fft.fft(temp_z)

        half_spec_x = spec_x[:(nop_fft / 2)]
        half_spec_y = spec_y[:(nop_fft / 2)]
        half_spec_z = spec_z[:(nop_fft / 2)]

        xs = np.roll(xs, -step_length)
        ys = np.roll(ys, -step_length)
        zs = np.roll(zs, -step_length)

        # CALCULATION OF THE SPECTRAL MATRIX
        spec_mat = np.zeros([int(nop_fft / 2), 3, 3])
        spec_mat[:, 0, 0] = half_spec_x * np.conj(half_spec_x)
        spec_mat[:, 1, 0] = half_spec_x * np.conj(half_spec_y)
        spec_mat[:, 2, 0] = half_spec_x * np.conj(half_spec_z)
        spec_mat[:, 0, 1] = half_spec_y * np.conj(half_spec_x)
        spec_mat[:, 1, 1] = half_spec_y * np.conj(half_spec_y)
        spec_mat[:, 2, 1] = half_spec_y * np.conj(half_spec_z)
        spec_mat[:, 0, 2] = half_spec_z * np.conj(half_spec_x)
        spec_mat[:, 1, 2] = half_spec_z * np.conj(half_spec_y)
        spec_mat[:, 2, 2] = half_spec_z * np.conj(half_spec_z)

        # Calculation of smoothed spectral matrix
        e_spec_mat = np.nan * np.ones(spec_mat.shape)

        off_idx = int((n_bin - 1) / 2)

        for k in range(off_idx, int(nop_fft / 2) - off_idx):
            for ir in range(3):
                for ic in range(3):
                    e_spec_mat[k, ir, ic] = np.sum(
                        aa[:n_bin] * spec_mat[(k - off_idx):(k + off_idx), ir, ic])

        # Calculation of the minimum variance direction and wave normal angle
        aaa2 = np.imag(e_spec_mat[:, 0, 1]) ** 2
        aaa2 += np.imag(e_spec_mat[:, 0, 2]) ** 2
        aaa2 += np.imag(e_spec_mat[:, 1, 2]) ** 2
        aaa2 = np.sqrt(aaa2[j, :])

        wn_x = -np.abs(np.imag(e_spec_mat[:, 1, 2]) / aaa2)
        wn_y = -np.abs(np.imag(e_spec_mat[:, 0, 2]) / aaa2)
        wn_z = np.imag(e_spec_mat[:, 0, 1]) / aaa2

        wave_angle[j, :] = np.arctan(np.sqrt(wn_x ** 2 + wn_y ** 2) / np.abs(wn_z))

        # CALCULATION OF THE DEGREE OF POLARISATION
        # calc of square of smoothed spec matrix
        for ir in range(3):
            for ic in range(3):
                sqrd_mat[:, ir, ic] = e_spec_mat[:, ir, 0] * e_spec_mat[:, 0, ic]
                sqrd_mat[:, ir, ic] += e_spec_mat[:, ir, 1] * e_spec_mat[:, 1, ic]
                sqrd_mat[:, ir, ic] += e_spec_mat[:, ir, 2] * e_spec_mat[:, 2, ic]

        trace_sqrd_mat = sqrd_mat[:, 0, 0] + sqrd_mat[:, 1, 1] + sqrd_mat[:, 2, 2]
        trace_spec_mat[j, :] = e_spec_mat[:, 0, 0] + e_spec_mat[:, 1, 1] + e_spec_mat[:, 2, 2]

        deg_pol[j, :] = trace_spec_mat[j, :] * np.nan
        deg_pol[j, off_idx:int(nop_fft / 2) - off_idx] = 3 * trace_sqrd_mat[
                                                             off_idx:int(nop_fft / 2) - off_idx]
        deg_pol[j, off_idx:int(nop_fft / 2) - off_idx] -= trace_spec_mat[j,
                                                          off_idx:int(nop_fft / 2) - off_idx] ** 2
        deg_pol[j, off_idx:int(nop_fft / 2) - off_idx] /= 2 * trace_spec_mat[j, off_idx:int(
            nop_fft / 2) - off_idx] ** 2

        # Calculation of helicity, ellipticity and the wave state vector
        alpha_x = np.sqrt(e_spec_mat[:, 0, 0])
        alpha_y = np.sqrt(e_spec_mat[:, 1, 1])
        alpha_z = np.sqrt(e_spec_mat[:, 2, 2])

        alpha_cos1_x = np.real(e_spec_mat[:, 0, 1]) / np.sqrt(e_spec_mat[:, 0, 0])
        alpha_sin1_x = -np.imag(e_spec_mat[j, :, 0, 1]) / np.sqrt(e_spec_mat[:, 0, 0])
        alpha_cos2_x = np.real(e_spec_mat[:, 0, 2]) / np.sqrt(e_spec_mat[:, 0, 0])
        alpha_sin2_x = -np.imag(e_spec_mat[j, :, 0, 2]) / np.sqrt(e_spec_mat[:, 0, 0])

        alpha_cos1_y = np.real(e_spec_mat[:, 1, 0]) / np.sqrt(e_spec_mat[:, 1, 1])
        alpha_sin1_y = -np.imag(e_spec_mat[:, 1, 0]) / np.sqrt(e_spec_mat[:, 1, 1])
        alpha_cos2_y = np.real(e_spec_mat[:, 1, 2]) / np.sqrt(e_spec_mat[:, 1, 1])
        alpha_sin2_y = -np.imag(e_spec_mat[:, 1, 2]) / np.sqrt(e_spec_mat[:, 1, 1])

        alpha_cos1_z = np.real(e_spec_mat[:, 2, 0]) / np.sqrt(e_spec_mat[:, 2, 2])
        alpha_sin1_z = -np.imag(e_spec_mat[:, 2, 0]) / np.sqrt(e_spec_mat[:, 2, 2])
        alpha_cos2_z = np.real(e_spec_mat[:, 2, 1]) / np.sqrt(e_spec_mat[:, 2, 2])
        alpha_sin2_z = -np.imag(e_spec_mat[:, 2, 1]) / np.sqrt(e_spec_mat[:, 2, 2])

        lambda_u[:, 0, 0] = alpha_x
        lambda_u[:, 1, 0] = alpha_y
        lambda_u[:, 2, 0] = alpha_z

        lambda_u[:, 0, 1] = np.complex(alpha_cos1_x, alpha_sin1_x)
        lambda_u[:, 0, 2] = np.complex(alpha_cos2_x, alpha_sin2_x)

        lambda_u[:, 1, 1] = np.complex(alpha_cos1_y, alpha_sin1_y)
        lambda_u[:, 1, 2] = np.complex(alpha_cos2_y, alpha_sin2_y)

        lambda_u[:, 2, 1] = np.complex(alpha_cos1_z, alpha_sin1_z)
        lambda_u[:, 2, 2] = np.complex(alpha_cos2_z, alpha_sin2_z)

        for k in range(int(nop_fft / 2)):
            for xyz in range(3):
                # HELICITY CALCULATION
                upper = np.sum(2 * np.real(lambda_u[k, xyz, :3]) * (np.imag(lambda_u[k, xyz, :3])))
                lower = np.sum(
                    (np.real(lambda_u[k, xyz, :3])) ** 2 - (np.imag(lambda_u[k, xyz, :3])) ** 2)

                if upper[j, k] > 0:
                    gamma = np.atan(upper[j, k] / lower[j, k])
                else:
                    gamma = np.pi + (np.pi + np.atan(upper[j, k] / lower[j, k]))

                lambda_u[k, xyz, :] = np.exp(np.complex(0, -0.5 * gamma)) * lambda_u[k, xyz, :]
                helic[j, k, xyz] = np.sqrt(np.sum(np.real(lambda_u[k, xyz, :3]) ** 2))
                helic[j, k, xyz] /= np.sqrt(np.sum(np.imag(lambda_u[k, xyz, :3]) ** 2))
                helic[j, k, xyz] = np.divide(1, helic[j, k, xyz])

                # ELLIPTICITY CALCULATION
                upper_e = np.sum(np.imag(lambda_u[k, xyz, :3]) * np.real(lambda_u[k, xyz, :3]))
                lower_e = np.sum(np.real(lambda_u[k, xyz, :2]) ** 2) - np.sum(
                    np.imag(lambda_u[k, xyz, :2]) ** 2)

                if upper_e > 0:
                    gamma_rot = np.arctan(upper_e / lower_e)
                else:
                    gamma_rot = np.pi + np.pi + np.atan(upper_e / lower_e)

                lam = lambda_u[k, xyz, :2]
                lambda_u_rot = np.exp(np.complex(0, -0.5 * gamma_rot)) * lam

                ellip[j, k, xyz] = np.sqrt(np.sum(np.imag(lambda_u_rot) ** 2))
                ellip[j, k, xyz] /= np.sqrt(np.sum(np.real(lambda_u_rot) ** 2))
                ellip[j, k, xyz] *= -(np.imag(e_spec_mat[k, 0, 1]) * np.sin(wave_angle[j, k]))
                ellip[j, k, xyz] /= np.abs(np.imag(e_spec_mat[k, 0, 1]) * np.sin(wave_angle[j, k]))

    # AVERAGING HELICITY AND ELLIPTICITY RESULTS
    ellipticity = np.mean(ellip, axis=-1)
    helicity = np.mean(helic, axis=-1)

    # CREATING OUTPUT PARAMETER
    time_line = ct[0] + np.abs(nop_fft / 2) / sample_freq \
                + np.arange(1, n_stp + 1) * step_length / sample_freq
    bin_width = sample_freq / nop_fft
    freq_line = bin_width * np.arange(1, nop_fft / 2 + 1)

    # scaling power results to units with meaning
    W = nop_fft * np.sum(smooth ** 2)

    power_spec = np.zeros([n_stp, int(nop_fft / 2)])
    power_spec[:, 1:nop_fft / 2 - 1] = 1 / W * 2 * trace_spec_mat[:, 1:nop_fft / 2 - 1] / bin_width
    power_spec[:, 1] = 1 / W * trace_spec_mat[:, 0] / bin_width
    power_spec[:, nop_fft / 2] = 1 / W * trace_spec_mat[:, nop_fft / 2] / bin_width

    # KICK OUT THE ANALYSIS OF THE WEAK SIGNALS
    wave_angle[power_spec < min_psd] = np.nan
    deg_pol[power_spec < min_psd] = np.nan
    ellipticity[power_spec < min_psd] = np.nan
    helicity[power_spec < min_psd] = np.nan

    # Save as DataArrays
    b_psd = xr.DataArray(power_spec, coords=[time_line, freq_line], dims=["t", "f"])
    wave_angle = xr.DataArray(wave_angle * 180 / np.pi, coords=[time_line, freq_line],
                              dims=["t", "f"])
    ellipticity = xr.DataArray(ellipticity, coords=[time_line, freq_line], dims=["t", "f"])
    deg_pol = xr.DataArray(deg_pol, coords=[time_line, freq_line], dims=["t", "f"])
    helicity = xr.DataArray(helicity, coords=[time_line, freq_line], dims=["t", "f"])

    b_psd.f.attrs["units"] = "Hz"
    wave_angle.f.attrs["units"] = "Hz"
    deg_pol.f.attrs["units"] = "Hz"
    ellipticity.f.attrs["units"] = "Hz"
    helicity.f.attrs["units"] = "Hz"

    return b_psd, wave_angle, deg_pol, ellipticity, helicity



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
vdf_to_deflux.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy import constants


def vdf_to_deflux(vdf=None):
    """Changes units to differential energy flux.

    Parameters
    ----------
    vdf : xarray.Dataset
        Time series of the 3D velocity distribution with :
            * time : Time samples.
            * data : 3D velocity distribution.
            * energy : Energy levels.
            * phi : Azimuthal angles.
            * theta : Elevation angle.

    Returns
    -------
    out : xarray.Dataset
        Time series of the 3D density energy flux with :
            * time : Time samples.
            * data : 3D density energy flux.
            * energy : Energy levels.
            * phi : Azimuthal angles.
            * theta : Elevation angle.

    """

    assert vdf is not None and isinstance(vdf, xr.Dataset)

    if vdf.attrs["species"] in ["ions", "i"]:
        mm = 1
    elif vdf.attrs["species"] in ["electrons", "e"]:
        mm = constants.m_e.value / constants.m_p.value
    else:
        raise ValueError("Invalid specie")

    if vdf.attrs["UNITS"].lower() == "s^3/cm^6":
        tmp_data = vdf.data.data * 1e30 / (1e6 * 0.53707 * mm ** 2)
    elif vdf.attrs["UNITS"].lower() == "s^3/m^6":
        tmp_data = vdf.data.data * 1e18 / (1e6 * 0.53707 * mm ** 2)
    elif vdf.attrs["UNITS"].lower() == "s^3/km^6":
        tmp_data = vdf.data.data / (1e6 * 0.53707 * mm ** 2)
    else:
        raise ValueError("Invalid unit")

    energy = vdf.energy.data

    if tmp_data.ndim == 2:
        tmp_data = tmp_data[:, :, None, None]

    data_r = np.reshape(tmp_data,
                        (tmp_data.shape[0], tmp_data.shape[1], np.prod(tmp_data.shape[2:])))

    if energy.ndim == 1:
        energy_mat = np.tile(energy, (len(vdf.time), np.prod(tmp_data.shape[2:]), 1))
        energy_mat = np.transpose(energy_mat, [0, 2, 1])
    elif energy.ndim == 2:
        energy_mat = np.tile(energy, (np.prod(tmp_data.shape[2:]), 1, 1))
        energy_mat = np.transpose(energy_mat, [1, 2, 0])
    else:
        raise ValueError("Invalid energy shape")

    data_r *= energy_mat ** 2
    tmp_data = np.reshape(data_r, tmp_data.shape)

    out = vdf.copy()
    out.data.data = np.squeeze(tmp_data)
    out.attrs["UNITS"] = "keV/(cm^2 s sr keV)"

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
ts_vec_xyz.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def ts_vec_xyz(t=None, data=None, attrs=None):
    """Create a time series containing a 1st order tensor.

    Parameters
    ----------
    t : numpy.ndarray
        Array of times.

    data : numpy.ndarray
        Data corresponding to the time list.

    attrs : dict
        Attributes of the data list.

    Returns
    -------
    out : xarray.DataArray
        1st order tensor time series.

    """

    assert t is not None and isinstance(t, np.ndarray)
    assert data is not None and isinstance(data, np.ndarray)
    assert data.ndim == 2 and data.shape[1] == 3

    if len(t) != len(data):
        raise IndexError("Time and data must have the same length")

    flag_attrs = True

    if attrs is None:
        flag_attrs = False

    out = xr.DataArray(data, coords=[t[:], ["x", "y", "z"]], dims=["time", "comp"])

    if flag_attrs:
        out.attrs = attrs

        out.attrs["TENSOR_ORDER"] = 1
    else:
        out.attrs["TENSOR_ORDER"] = 1

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
fk_power_spectrum_4sc.py

@author : Louis RICHARD
"""

import bisect
import numpy as np
import xarray as xr

from ..pyrf.resample import resample
from ..pyrf.avg_4sc import avg_4sc
from ..pyrf.time_clip import time_clip
from ..pyrf.wavelet import wavelet


def fk_power_spectrum_4sc(e=None, r=None, b=None, tints=None, cav=8, num_k=500, num_f=200, df=None,
                          w_width=1, f_range=None):
    """Calculates the frequency-wave number power spectrum using the four MMS spacecraft. Uses a
    generalization of mms.fk_powerspectrum. Wavelet based cross-spectral analysis is used to
    calculate the phase difference each spacecraft pair and determine 3D wave vector. A
    generalization of the method used in mms.fk_powerspectrum to four point measurements.

    Parameters
    ----------
    e : list of xarray.DataArray
        Fields to apply 4SC cross-spectral analysis to. e.g., e or b fields (if multiple
        components only the first is used).

    r : list of xarray.DataArray
        Positions of the four spacecraft.

    b : list of xarray.DataArray
        background magnetic field in the same coordinates as r. Used to determine the parallel
        and perpendicular wave numbers using 4SC average.

    tints : list of str
        Time interval over which the power spectrum is calculated. To avoid boundary effects use
        a longer time interval for e and b.

    cav : int, optional
        Number of points in time series used to estimate phase. Default ``cav`` = 8.

    num_k : int, optional
        Number of wave numbers used in spectrogram. Default ``num_k`` = 500.

    df : float, optional
        Linear spacing of frequencies. Default ``df`` = None (log spacing).

    num_f : int, optional
        Number of frequencies used in spectrogram. Default ``num_f`` = 200.

    w_width : float, optional
        Multiplier for Morlet wavelet width. Default ``w_width`` = 1.

    f_range : list of float, optional
        Frequency range for k-k plots. [minf maxf].

    Returns
    -------
    out : xarray.Dataset
        Dataset of array of powers as a function of frequency and wavenumber. Power is normalized
        to the maximum value.

    Notes
    -----
    Wavelength must be larger than twice the spacecraft separations, otherwise spatial aliasing
    will occur.

    Examples
    --------
    >>> from pyrfu import mms
    >>> power = mms.fk_power_spectrum_4sc(e_par, r_xyz, b_xyz, tints)
    >>> power = mms.fk_power_spectrum_4sc(b_scmfac_x, r_xyz, b_xyz, tints, 4, 500, 2, 10, 2)
    >>> # Plot
    >>> import matplotlib.pyplot as plt
    >>> from pyrfu.plot import plot_spectr
    >>> fig, ax = plt.subplots(1)
    >>> ax, cax = plot_spectr(ax, power.kmagf, cscale="log", cmap="viridis")
    >>> ax.set_xlabel("$|k|$ [m$^{-1}$]")
    >>> ax.set_ylabel("$f$ [Hz]")

    """

    assert e is not None and isinstance(e, list) and len(e) == 4
    assert r is not None and isinstance(r, list) and len(r) == 4
    assert b is not None and isinstance(b, list) and len(b) == 4
    assert tints is not None and isinstance(tints, list)
    assert isinstance(tints[0], str) and isinstance(tints[1], str)

    for i in range(4):
        assert isinstance(e[i], xr.DataArray)
        assert isinstance(r[i], xr.DataArray)
        assert isinstance(b[i], xr.DataArray)

    ic = np.arange(1, 5)

    e = [resample(e[i - 1], e[0]) for i in ic]
    r = [resample(r[i - 1], e[0]) for i in ic]
    b = [resample(b[i - 1], e[0]) for i in ic]

    b_avg = avg_4sc(b)

    times = e[0].time
    use_linear = not df is None

    idx = time_clip(e[0].time, tints)

    # If odd, remove last data point (as is done in irf_wavelet)
    if len(idx) % 2:
        idx = idx[:-1]

    if use_linear:
        w = []
        for i in range(4):
            w.append(wavelet(e[i], linear=df, returnpower=False, wavelet_width=5.36 * w_width))
    else:
        w = []
        for i in range(4):
            w.append(wavelet(e[i], nf=num_f, returnpower=False, wavelet_width=5.36 * w_width))

    num_f = len(w[0].frequency)

    times = time_clip(times, tints)
    nt = len(times)

    w = [time_clip(w[i], tints) for i in range(4)]

    fk_power = 0
    for i in range(4):
        fk_power += w[i].data * np.conj(w[i].data) / 4

    n = int(np.floor(nt/cav)-1)
    pos_av = cav / 2 + np.arange(n) * cav
    av_times = times[pos_av.astype(int)]

    b_avg = resample(b_avg, av_times)

    r = [resample(r[i], av_times) for i in range(4)]

    cx12, cx13, cx14 = [np.zeros((n, num_f), dtype="complex128") for _ in range(3)]
    cx23, cx24, cx34 = [np.zeros((n, num_f), dtype="complex128") for _ in range(3)]

    power_avg = np.zeros((n, num_f), dtype="complex128")

    for m, pos_avm in enumerate(pos_av):
        lb, ub = [int(pos_avm - cav / 2 + 1), int(pos_avm + cav / 2)]

        cx12[m, :] = np.nanmean(w[0].data[lb:ub, :] * np.conj(w[1].data[lb:ub, :]), axis=0)
        cx13[m, :] = np.nanmean(w[0].data[lb:ub, :] * np.conj(w[2].data[lb:ub, :]), axis=0)
        cx14[m, :] = np.nanmean(w[0].data[lb:ub, :] * np.conj(w[3].data[lb:ub, :]), axis=0)
        cx23[m, :] = np.nanmean(w[1].data[lb:ub, :] * np.conj(w[2].data[lb:ub, :]), axis=0)
        cx24[m, :] = np.nanmean(w[1].data[lb:ub, :] * np.conj(w[3].data[lb:ub, :]), axis=0)
        cx34[m, :] = np.nanmean(w[2].data[lb:ub, :] * np.conj(w[3].data[lb:ub, :]), axis=0)

        power_avg[m, :] = np.nanmean(fk_power[lb:ub, :], axis=0)

    # Compute phase differences between each spacecraft pair
    th12 = np.arctan2(np.imag(cx12), np.real(cx12))
    th13 = np.arctan2(np.imag(cx13), np.real(cx13))
    th14 = np.arctan2(np.imag(cx14), np.real(cx14))
    th23 = np.arctan2(np.imag(cx23), np.real(cx23))
    th24 = np.arctan2(np.imag(cx24), np.real(cx24))
    th34 = np.arctan2(np.imag(cx34), np.real(cx34))

    w_mat = 2 * np.pi * np.tile(w[0].frequency.data, (n, 1))

    # Convert phase difference to time delay
    dt12, dt13, dt14, dt23, dt24, dt34 = [th / w_mat for th in [th12, th13, th14, th23, th24, th34]]

    # Weighted averaged time delay using all spacecraft pairs
    dt2 = 0.5 * dt12 + 0.2 * (dt13 - dt23) + 0.2 * (dt14 - dt24) + 0.1 * (dt14 - dt34 - dt23)
    dt3 = 0.5 * dt13 + 0.2 * (dt12 + dt23) + 0.2 * (dt14 - dt34) + 0.1 * (dt12 + dt24 - dt34)
    dt4 = 0.5 * dt14 + 0.2 * (dt12 + dt24) + 0.2 * (dt13 + dt34) + 0.1 * (dt12 + dt23 + dt34)

    # Compute phase speeds
    r = [r[i].data for i in range(4)]

    k_x, k_y, k_z = [np.zeros((n, num_f)) for _ in range(3)]

    # Volumetric tensor with SC1 as center.
    dr = np.reshape(np.hstack(r[1:]), (n, 3, 3)) - np.reshape(np.tile(r[0], (1, 3)), (n, 3, 3))
    dr = np.transpose(dr, [0, 2, 1])

    # Delay tensor with SC1 as center.
    # dT = np.reshape(np.hstack([dt2,dt3,dt4]),(N,num_f,3))
    tau = np.dstack([dt2, dt3, dt4])

    for ii in range(num_f):
        m = np.linalg.solve(dr, np.squeeze(tau[:, ii, :]))

        k_x[:, ii] = 2 * np.pi * w[0].frequency[ii].data * m[:, 0]
        k_y[:, ii] = 2 * np.pi * w[0].frequency[ii].data * m[:, 1]
        k_z[:, ii] = 2 * np.pi * w[0].frequency[ii].data * m[:, 2]

    k_x, k_y, k_z = [k / 1e3 for k in [k_x, k_y, k_z]]

    k_mag = np.linalg.norm(np.array([k_x, k_y, k_z]), axis=0)

    b_avg_x_mat = np.tile(b_avg.data[:, 0], (num_f, 1)).T
    b_avg_y_mat = np.tile(b_avg.data[:, 1], (num_f, 1)).T
    b_avg_z_mat = np.tile(b_avg.data[:, 2], (num_f, 1)).T

    b_avg_abs = np.linalg.norm(b_avg, axis=1)
    b_avg_abs_mat = np.tile(b_avg_abs, (num_f, 1)).T

    k_par = (k_x * b_avg_x_mat + k_y * b_avg_y_mat + k_z * b_avg_z_mat) / b_avg_abs_mat
    k_perp = np.sqrt(k_mag ** 2 - k_par ** 2)

    k_max = np.max(k_mag) * 1.1
    k_min = -k_max
    k_vec = np.linspace(-k_max, k_max, num_k)
    k_mag_vec = np.linspace(0, k_max, num_k)

    dk_mag = k_max / num_k
    dk = 2 * k_max / num_k

    # Sort power into frequency and wave vector
    print("notice : Computing power versus kx,f; ky,f, kz,f")
    power_k_x_f, power_k_y_f, power_k_z_f = [np.zeros((num_f, num_k)) for _ in range(3)]
    power_k_mag_f = np.zeros((num_f, num_k))

    for nn in range(num_f):
        k_x_number = np.floor((k_x[:, nn] - k_min) / dk).astype(int)
        k_y_number = np.floor((k_y[:, nn] - k_min) / dk).astype(int)
        k_z_number = np.floor((k_z[:, nn] - k_min) / dk).astype(int)
        k_number = np.floor((k_mag[:, nn]) / dk_mag).astype(int)

        power_k_x_f[nn, k_x_number] += np.real(power_avg[:, nn])
        power_k_y_f[nn, k_y_number] += np.real(power_avg[:, nn])
        power_k_z_f[nn, k_z_number] += np.real(power_avg[:, nn])

        power_k_mag_f[nn, k_number] += np.real(power_avg[:, nn])

    # power_k_x_f[power_k_x_f == 0] 	= np.nan
    # power_k_y_f[power_k_y_f == 0] 	= np.nan
    # power_k_z_f[power_k_z_f == 0] 	= np.nan
    # power_k_mag_f[power_k_mag_f == 0] = np.nan

    power_k_x_f /= np.max(power_k_x_f)
    power_k_y_f /= np.max(power_k_y_f)
    power_k_z_f /= np.max(power_k_z_f)
    power_k_mag_f /= np.max(power_k_mag_f)

    # power_k_x_f[power_k_x_f < 1.0e-6] 		= 1e-6
    # power_k_y_f[power_k_y_f < 1.0e-6] 		= 1e-6
    # power_k_z_f[power_k_z_f < 1.0e-6] 		= 1e-6
    # power_k_mag_f[power_k_mag_f < 1.0e-6] 	= 1e-6

    frequencies = w[0].frequency.data
    idx_f = np.arange(num_f)

    if f_range is not None:
        idx_min_freq = bisect.bisect_left(np.min(f_range))
        idx_max_freq = bisect.bisect_left(np.max(f_range))
        idx_f = idx_f[idx_min_freq:idx_max_freq]

    print("notice : Computing power versus kx,ky; kx,kz; ky,kz\n")
    power_k_x_k_y = np.zeros((num_k, num_k))
    power_k_x_k_z = np.zeros((num_k, num_k))
    power_k_y_k_z = np.zeros((num_k, num_k))
    power_k_perp_k_par = np.zeros((num_k, num_k))

    for nn in idx_f:
        k_x_number = np.floor((k_x[:, nn] - k_min) / dk).astype(int)
        k_y_number = np.floor((k_y[:, nn] - k_min) / dk).astype(int)
        k_z_number = np.floor((k_z[:, nn] - k_min) / dk).astype(int)

        k_par_number = np.floor((k_par[:, nn] - k_min) / dk).astype(int)
        k_perp_number = np.floor((k_perp[:, nn]) / dk_mag).astype(int)

        power_k_x_k_y[k_y_number, k_x_number] += np.real(power_avg[:, nn])
        power_k_x_k_z[k_z_number, k_x_number] += np.real(power_avg[:, nn])
        power_k_y_k_z[k_z_number, k_y_number] += np.real(power_avg[:, nn])

        power_k_perp_k_par[k_par_number, k_perp_number] += np.real(power_avg[:, nn])

    # power_k_x_k_y[power_k_x_k_y == 0] = np.nan
    # power_k_x_k_z[power_k_x_k_z == 0] = np.nan
    # power_k_y_k_z[power_k_y_k_z == 0] = np.nan
    # power_k_perp_k_par[power_k_perp_k_par == 0] = np.nan

    power_k_x_k_y /= np.max(power_k_x_k_y)
    power_k_x_k_z /= np.max(power_k_x_k_z)
    power_k_y_k_z /= np.max(power_k_y_k_z)
    power_k_perp_k_par /= np.max(power_k_perp_k_par)

    # power_k_x_k_y(power_k_x_k_y < 1.0e-6) 				= 1e-6
    # power_k_x_k_z(power_k_x_k_z < 1.0e-6) 				= 1e-6
    # power_k_y_k_z(power_k_y_k_z < 1.0e-6) 				= 1e-6
    # power_k_perp_k_par[power_k_perp_k_par < 1.0e-6] 	= 1e-6

    out_dict = {"k_x_f": (["k_x", "f"], power_k_x_f.T), "k_y_f": (["k_x", "f"], power_k_y_f.T),
                "k_z_f": (["k_x", "f"], power_k_z_f.T),
                "k_mag_f": (["k_mag", "f"], power_k_mag_f.T),
                "k_x_k_y": (["k_x", "k_y"], power_k_x_k_y.T),
                "k_x_k_z": (["kx", "kz"], power_k_x_k_z.T),
                "k_y_k_z": (["k_y", "k_z"], power_k_y_k_z.T),
                "k_perp_k_par": (["k_perp", "k_par"], power_k_perp_k_par.T),
                "k_x": k_vec, "k_y": k_vec, "k_z": k_vec, "k_mag": k_mag_vec, "k_perp": k_mag_vec,
                "k_par": k_vec,
                "f": frequencies}

    out = xr.Dataset(out_dict)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
estimate_phase_speed.py

@author : Louis RICHARD
"""

import numpy as np

from scipy.optimize import curve_fit


def estimate_phase_speed(f_k_power=None, f=None, k=None, f_min=100.):
    """Simple function to estimate the phase speed from the frequency wave number power spectrum.
    Fits :math:`f = v k/ 2 \\pi` to the power spectrum.

    Parameters
    ----------
    f_k_power : numpy.ndarray
        2D array of powers.

    f : numpy.ndarray
        1D array of frequencies.

    k : numpy.ndarray
        1D array of wave numbers.

    f_min : float or int, optional
        Set low frequency threshold of points used to estimate the speed. Default ``f_min`` = 100

    Returns
    -------
    vph : float
        Estimated phase speed by fitting linear dispersion relation to data.

    Notes
    -----
    Draft version but seems to work well. Does not yet handle multiple modes in the same power
    spectrum.

    See also
    --------
    pyrfu.mms.fk_power_spectrum : Calculates the frequency-wave number power spectrum
    pyrfu.mms.probe_align_times : Returns times when f-a electrostatic waves can be characterized.

    """

    assert f_k_power is not None and isinstance(f_k_power, np.ndarray)
    assert f is not None and isinstance(f, np.ndarray)
    assert k is not None and isinstance(k, np.ndarray)
    assert isinstance(f_min, float)

    # Remove spurious points; specifically at large k.
    k_max = 2.0 * np.max(k) / 3.0
    power_temp = f_k_power
    rm_k = np.where(abs(k) > k_max)
    rm_f = np.where(f < f_min)

    power_temp[:, rm_k] = 0.0
    power_temp[rm_f, :] = 0.0
    power_temp[np.isnan(power_temp)] = 0.0

    # Find position of maximum power to guess vph
    [k_mat, f_mat] = np.meshgrid(k, f)
    max_pos = np.unravel_index(np.argmax(power_temp), power_temp.shape)

    max_f, max_k = [f_mat[max_pos], k_mat[max_pos]]

    # Initial guess
    vph_guess = max_f / max_k

    if vph_guess > 0.0:
        power_temp[:, k < 0.0] = 0
    else:
        power_temp[:, k > 0.0] = 0

    vph_range = [vph_guess / 3, vph_guess * 3]

    # Find all points around this guess vph
    highppos = np.where(power_temp > 0.3 * np.max(power_temp))

    p_power, f_power, k_power = [power_temp[highppos], f_mat[highppos], k_mat[highppos]]

    p_power2 = []
    f_power2 = []
    k_power2 = []

    for ii, pp in enumerate(p_power):
        if np.abs(vph_range[0]) < np.abs(f_power[ii] / k_power[ii]) < np.abs(vph_range[1]):
            p_power2.append(pp)
            f_power2.append(f_power[ii])
            k_power2.append(k_power[ii])

    p_power2, f_power2, k_power2 = [np.array(p_power2), np.array(f_power2), np.array(k_power2)]

    weights = 1 + np.log10(p_power2 / np.max(p_power2))

    def fun(x, a):
        return a * x

    popt, _ = curve_fit(fun, k_power2, f_power2, p0=vph_guess, sigma=weights)
    vph = popt[0] * 2 * np.pi

    return vph



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
list_files.py

@author : Louis RICHARD
"""

import os
import re
import bisect
import datetime

from dateutil import parser
from dateutil.rrule import rrule, DAILY

from .mms_config import CONFIG


def list_files(tint=None, mms_id="1", var=None):
    """Find files in the data directories of the target instrument, data type, data rate, mms_id and
    level during the target time interval

    Parameters
    ----------
    tint : list
        Time interval

    mms_id : str or int
        Index of the spacecraft

    var : dict
        Dictionary containing 4 keys
            * var["inst"] : name of the instrument
            * var["tmmode"] : data rate
            * var["lev"] : data level
            * var["dtype"] : data type

    Returns
    -------
    files : list
        List of files corresponding to the parameters in the selected time interval

    """

    data_path = CONFIG["local_data_dir"]

    if var is None:
        raise ValueError("var is empty")

    files_out = []

    if not isinstance(mms_id, str):
        mms_id = str(mms_id)
    # directory and file name search patterns
    #   -assume directories are of the form:
    #      (srvy, SITL): spacecraft/instrument/rate/level[/datatype]/year/month/
    #      (brst): spacecraft/instrument/rate/level[/datatype]/year/month/day/
    #   -assume file names are of the form:
    #      spacecraft_instrument_rate_level[_datatype]_YYYYMMDD[hhmmss]_version.cdf

    file_name = "mms" + mms_id + "_" + var["inst"] + "_" + var["tmmode"] + "_" + var[
        "lev"] + "(_)?.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf"

    days = rrule(DAILY, dtstart=parser.parse(parser.parse(tint[0]).strftime("%Y-%m-%d")),
                 until=parser.parse(tint[1]) - datetime.timedelta(seconds=1))

    if var["dtype"] == "" or var["dtype"] is None:
        level_and_dtype = var["lev"]
    else:
        level_and_dtype = os.sep.join([var["lev"], var["dtype"]])

    for date in days:
        if var["tmmode"] == "brst":
            local_dir = os.sep.join(
                [data_path, f"mms{mms_id}", var["inst"], var["tmmode"], level_and_dtype,
                 date.strftime("%Y"), date.strftime("%m"), date.strftime("%d")])
        else:
            local_dir = os.sep.join(
                [data_path, f"mms{mms_id}", var["inst"], var["tmmode"], level_and_dtype,
                 date.strftime("%Y"), date.strftime("%m")])

        if os.name == "nt":
            full_path = os.sep.join([re.escape(local_dir)+os.sep, file_name])
        else:
            full_path = os.sep.join([re.escape(local_dir), file_name])

        regex = re.compile(full_path)

        for root, dirs, files in os.walk(local_dir):
            for file in files:
                this_file = os.sep.join([root, file])

                matches = regex.match(this_file)
                if matches:
                    this_time = parser.parse(matches.groups()[1])
                    if (parser.parse(parser.parse(tint[0]).strftime("%Y-%m-%d")) <= this_time
                            <= parser.parse(tint[1]) - datetime.timedelta(seconds=1)):
                        if this_file not in files_out:
                            files_out.append(
                                {"file_name": file, "timetag": "", "full_name": this_file,
                                 "file_size": ""})

    in_files = files_out

    file_name = "mms.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf"

    file_times = []

    regex = re.compile(file_name)

    for file in in_files:
        matches = regex.match(file["file_name"])
        if matches:
            file_times.append((file["file_name"], parser.parse(matches.groups()[0]).timestamp(),
                               file["timetag"], file["file_size"]))

    # sort in time
    sorted_files = sorted(file_times, key=lambda x: x[1])

    times = [t[1] for t in sorted_files]

    idx_min = bisect.bisect_left(times, parser.parse(tint[0]).timestamp())

    # note: purposefully liberal here; include one extra file so that we always get the burst mode
    # data
    if idx_min == 0:
        files_in_interval = []
        for f in sorted_files[idx_min:]:
            files_in_interval.append({"file_name": f[0], "timetag": f[2], "file_size": f[3]})
    else:
        files_in_interval = []
        for f in sorted_files[idx_min - 1:]:
            files_in_interval.append({"file_name": f[0], "timetag": f[2], "file_size": f[3]})

    local_files = []

    file_names = [f["file_name"] for f in files_in_interval]

    for file in files_out:
        if file["file_name"] in file_names:
            local_files.append(file["full_name"])

    return sorted(local_files)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
wavelet.py

@author : Louis RICHARD
"""

import multiprocessing as mp
import numpy as np
import xarray as xr
import pyfftw
import matplotlib.pyplot as plt

from astropy.time import Time


def wavelet(inp=None, **kwargs):
    """Computes wavelet spectrogram based on fast FFT algorithm.

    Parameters
    ----------
    inp : xarray.DataArray
        Input quantity.

    **kwargs : dict
        Hash table of keyword arguments with :
            * fs : int or float
                Sampling frequency of the input time series.

            * f : list or numpy.ndarray
                Vector [f_min f_max], calculate spectra between frequencies f_min and f_max.

            * nf : int or float
                Number of frequency bins.

            * wavelet_width : int or float
                Width of the Morlet wavelet. Default 5.36.

            * linear : float
                Linear spacing between frequencies of df.

            * return_power : bool
                Set to True to return the power, False for complex wavelet transform. Default True.

            * cut_edge : bool
                Set to True to set points affected by edge effects to NaN, False to keep edge
                affect points. Default True

    Returns
    -------
    out : xarray.DataArray or xarray.Dataset
        Wavelet transform of the input.

    """

    # Default values
    # Fs
    assert inp is not None and isinstance(inp, xr.DataArray)

    # Time bounds
    start_time = Time(inp.time.data[0], format="datetime64").unix
    end_time = Time(inp.time.data[-1], format="datetime64").unix

    # Time interval
    tint = end_time - start_time

    # Sampling frequency
    fs = len(inp) / tint

    # Unpack time and data
    t, data = [inp.time.data.view("i8") * 1e-9, inp.data]

    # f
    a_min, a_max = [0.01, 2]
    f_min, f_max = [.5 * fs / 10 ** a_max, .5 * fs / 10 ** a_min]

    # nf
    nf = 200

    # wavelet_width
    wavelet_width, delta_f = [5.36, 100]

    return_power, cut_edge, linear_df, plot_flag = [True, True, False, False]

    if kwargs.get("return_power"):
        return_power = kwargs["return_power"]

    if kwargs.get("cut_edge"):
        cut_edge = kwargs["cut_edge"]

    if kwargs.get("fs"):
        if isinstance(kwargs["fs"], (int, float)):
            fs = kwargs["fs"]
        else:
            raise TypeError("fs must be numeric")

    if kwargs.get("nf"):
        if isinstance(kwargs["nf"], (int, float)):
            nf = kwargs["nf"]
        else:
            raise TypeError("nf must be numeric")

    if kwargs.get("linear"):
        linear_df = True
        if isinstance(kwargs["linear"], (int, float)):
            delta_f = kwargs["linear"]
        else:
            raise Warning("Unknown input for linear delta_f set to 100")

    if kwargs.get("wavelet_width"):
        if isinstance(kwargs["wavelet_width"], (int, float)):
            wavelet_width = kwargs["wavelet_width"]
        else:
            raise TypeError("wavelet_width must be numeric")

    if kwargs.get("f"):
        if isinstance(kwargs["f"], (np.ndarray, list)):
            if len(kwargs["f"]) == 2:
                f_min = kwargs["f"][0]
                f_max = kwargs["f"][1]
            else:
                raise IndexError("f should have vector with 2 elements as parameter value")
        else:
            raise TypeError("f must be a list or array")

    if kwargs.get("plot"):
        plot_flag = kwargs["plot"]

    w0, a_number, sigma = [fs / 2, nf, wavelet_width / (fs / 2)]

    if linear_df:
        a_number = np.floor(w0 / delta_f).astype(int)

        f_min, f_max = [delta_f, a_number * delta_f]

        a = w0 / (np.linspace(f_max, f_min, a_number))
    else:
        a_min, a_max = [np.log10(.5 * fs / f_max), np.log10(.5 * fs / f_min)]

        a = np.logspace(a_min, a_max, a_number)

    # Remove the last sample if the total number of samples is odd
    if len(data) / 2 != np.floor(len(data) / 2):
        t, data = [t[:-1], data[:-1, ...]]

    # Check for NaNs
    a[np.isnan(a)] = 0

    # Find the frequencies for an FFT of all data
    nd2, nyq = [len(data) / 2, 1 / 2]

    freq = fs * np.arange(1, nd2 + 1) / nd2 * nyq

    # The frequencies corresponding to FFT
    w = np.hstack([0, freq, -np.flip(freq[:-1])])

    # Get the correct frequencies for the wavelet transform
    new_freq = w0 / a

    if len(inp.shape) == 1:
        out_dict, power2 = [None, np.zeros((len(inp.data), nf))]
    elif len(inp.shape) == 2:
        out_dict, power2 = [{}, None]
    else:
        raise TypeError("Invalid shape of the inp")

    new_freq_mat, _ = np.meshgrid(new_freq, w)

    _, ww = np.meshgrid(a, w)  # Matrix form

    # if scalar add virtual axis
    if len(inp.shape) == 1:
        data = data[:, np.newaxis]

    # go through all the data columns
    for i in range(data.shape[1]):
        # Make the FFT of all data
        data_col = data[:, i]

        # Wavelet transform of the data
        # Forward FFT
        s_w = pyfftw.interfaces.numpy_fft.fft(data_col, threads=mp.cpu_count())

        aa, s_ww = np.meshgrid(a, s_w)  # Matrix form

        # Calculate the FFT of the wavelet transform
        w_w = np.sqrt(1) * s_ww * np.exp(-sigma * sigma * ((aa * ww - w0) ** 2) / 2)

        # Backward FFT
        power = pyfftw.interfaces.numpy_fft.ifft(w_w, axis=0, threads=mp.cpu_count())

        # Calculate the power spectrum
        if return_power:
            power = np.absolute((2 * np.pi) * np.conj(power) * power / new_freq_mat)
        else:
            power = np.sqrt(np.absolute((2 * np.pi) / new_freq_mat)) * power

        # Remove data possibly influenced by edge effects
        power2 = power

        if cut_edge:
            censure = np.floor(2 * a).astype(int)

            for j in range(a_number):
                power2[1:censure[j], j] = np.nan

                power2[len(data_col) - censure[j]:len(data_col), j] = np.nan
        else:
            continue

        if len(inp.shape) == 2:
            out_dict[inp.comp.data[i]] = (["time", "frequency"], power2)
        else:
            continue

    if len(inp.shape) == 1:
        out = xr.DataArray(power2, coords=[Time(t, format="unix").datetime64, new_freq],
                           dims=["time", "frequency"])
    elif len(inp.shape) == 2:
        out = xr.Dataset(out_dict,
                         coords={"time": Time(t, format="unix").datetime64, "frequency": new_freq})
    else:
        raise TypeError("Invalid shape")

    if plot_flag:
        if isinstance(out, xr.Dataset):
            fig, axs = plt.subplots(3, sharex="all")
            fig.subplots_adjust(hspace=0)
            axs[0].pcolormesh(out.time, out.frequency, out.x.data, cmap="jet")
            axs[0].set_yscale('log')
            axs[0].set_ylabel("f [Hz]")
            axs[1].pcolormesh(out.time, out.frequency, out.y.data, cmap="jet")
            axs[1].set_yscale('log')
            axs[1].set_ylabel("f [Hz]")
            axs[2].pcolormesh(out.time, out.frequency, out.z.data, cmap="jet")
            axs[2].set_yscale('log')
            axs[2].set_ylabel("f [Hz]")
        else:
            fig, axs = plt.subplots(1)
            fig.subplots_adjust(hspace=0)
            axs.pcolormesh(out.time, out.frequency, out.data.T, cmap="jet")
            axs.set_yscale('log')
            axs.set_ylabel("f [Hz]")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
pvi.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def pvi(x=None, scale=10):
    """Returns the PVI of a time series.

    .. math::

        y = \\frac{|x_i - x_{i+s}|^2}{<|x_i - x_{i+s}|^2>}

    where :math:`s` is the scale.

    Parameters
    ----------
    x : xarray.DataArray
        Input time series.

    scale : int
        Scale at which to compute the PVI.

    Returns
    -------
    values : xarray.DataArray
        An xarray containing the pvi of the original time series.

    """

    assert x is not None and isinstance(x, xr.DataArray)
    assert isinstance(scale, int)

    data = x.data

    if len(data.shape) == 1:
        data = data[:, np.newaxis]
    else:
        pass

    f = np.abs((data[scale:, :] - data[:-scale, :]))
    f2 = np.sum(f ** 2, axis=1)
    sigma = np.mean(f2)
    result = np.array(f2 / sigma)

    t = x.dims[0]
    time = x.coords[t].data

    result = xr.DataArray(result, coords=[time[0:len(f)]], dims=[t], attrs=x.attrs)

    result.attrs["units"] = "dimensionless"

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
psd.py

@author : Louis RICHARD
"""

import warnings
import numpy as np
import xarray as xr

from scipy import signal


def psd(inp=None, n_fft=256, n_overlap=128, window="hamming", d_flag="constant", scaling="density"):
    """Estimate power spectral density using Welch's method.

    Welch's method [11]_ computes an estimate of the power spectral density by dividing the data
    into overlapping segments, computing a modified periodogram for each segment and averaging the
    periodograms.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of measurement values.

    n_fft : int, optional
        Length of the FFT used, if a zero padded FFT is desired. Default to 256.

    n_overlap : int, optional
        Number of points to overlap between segments. Default to 128.

    window : str, optional
        Desired window to use. It is passed to `get_window` to generate the window values,
        which are DFT-even by default. See "get_window" or a list of windows and required
        parameters. Default Hanning

    d_flag : str, optional
        Specifies how to detrend each segment. It is passed as the "type" argument to the
        "detrend" function. Default to "constant".

    scaling : str, optional
        Selects between computing the power spectral density ('density') where `Pxx` has units of
        V**2/Hz and computing the power spectrum ("spectrum") where "Pxx" has units of V**2,
        if `x` is measured in V and "fs" is measured in Hz. Default to 'density'

    Returns
    -------
    out : xarray.DataArray
        Power spectral density or power spectrum of inp.

    References
    ----------
    .. [11] P. Welch, "The use of the fast Fourier transform for the estimation of power spectra:
            A method based on time averaging over short, modified periodograms", IEEE Trans. Audio
            Electroacoust. vol. 15, pp. 70-73, 1967.

    """

    assert inp is not None and isinstance(inp, xr.DataArray)

    if inp.ndim == 2 and inp.shape[-1] == 3:
        inp = np.abs(inp)

    if n_overlap is None:
        n_persegs = 256
        n_overlap = n_persegs / 2
    else:
        n_persegs = 2 * n_overlap

    if n_fft < n_persegs:
        n_fft = n_persegs
        warnings.warn("nfft < n_persegs. set to n_persegs", UserWarning)

    dt = np.median(np.diff(inp.time.data)).astype(float) * 1e-9
    fs = 1 / dt

    f, p_xx = signal.welch(inp.data, nfft=n_fft, fs=fs, window=window, noverlap=n_overlap,
                           detrend=d_flag, nperseg=n_persegs, scaling=scaling, return_onesided=True,
                           axis=-1)

    out = xr.DataArray(p_xx, coords=[f], dims=["f"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
dynamic_press.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy import constants


def dynamic_press(n=None, v=None, s="i"):
    """Computes dynamic pressure.

    Parameters
    ----------
    n : xarray.DataArray
        Time series of the number density of the specie.

    v : xarray.DataArray
        Time series of the bulk velocity of the specie.

    s : {"i", "e"}, optional
        Specie. default "i".

    Returns
    -------
    p_dyn : xarray.DataArray
        Time series of the dynamic pressure of the specie.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load ion bulk velocity and remove spintone
    >>> v_xyz_i = mms.get_data("Vi_gse_fpi_fast_l2", tint, mms_id)
    >>> st_xyz_i = mms.get_data("STi_gse_fpi_fast_l2", tint, mms_id)
    >>> v_xyz_i = v_xyz_i - st_xyz_i
    >>> # Ion number density
    >>> n_i = mms.get_data("Ni_fpi_fast_l2", tint, mms_id)
    >>> # Compute dynamic pressure
    >>> p = pyrf.dynamic_press(n_i, v_xyz_i, s="i")

    """

    assert n is not None and isinstance(n, xr.DataArray)
    assert v is not None and isinstance(v, xr.DataArray)

    if s == "i":
        m = constants.m_p.value
    elif s == "e":
        m = constants.m_e.value
    else:
        raise ValueError("Unknown specie")

    p_dyn = n * m * np.linalg.norm(v, axis=0) ** 2

    return p_dyn



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
iso2unix.py

@author : Louis RICHARD
"""

import numpy as np

from astropy.time import Time


def iso2unix(t=None):
    """
    Converts time in iso format to unix

    Parameters
    ----------
    t : list of str
        Time.

    Returns
    -------
    out : list of float
        Time in unix format.

    """

    assert t is not None and isinstance(t, (list, np.ndarray))

    # Convert iso time to unix
    out = Time(t, format="iso").unix

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
date_str.py

@author : Louis RICHARD
"""

from dateutil import parser


def date_str(tint=None, fmt=1):
    """Creates a string corresponding to time interval for output plot naming.

    Parameters
    ----------
    tint : list of str
        Time interval.

    fmt : int
        Format of the output :
            * 1 : "%Y%m%d_%H%M"
            * 2 : "%y%m%d%H%M%S"
            * 3 : "%Y%m%d_%H%M%S"_"%H%M%S"
            * 4 : "%Y%m%d_%H%M%S"_"%Y%m%d_%H%M%S"

    Returns
    -------
    out : str
        String corresponding to the time interval in the desired format.

    """

    assert tint is not None and isinstance(tint, list) and len(tint) == 2

    t1 = parser.parse(tint[0])
    t2 = parser.parse(tint[1])

    if fmt == 1:
        out = t1.strftime("%Y%m%d_%H%M")
    elif fmt == 2:
        out = t1.strftime("%y%m%d%H%M%S")
    elif fmt == 3:
        out = "_".join([t1.strftime("%Y%m%d_%H%M%S"), t2.strftime("%H%M%S")])
    elif fmt == 4:
        out = "_".join([t1.strftime("%Y%m%d_%H%M%S"), t2.strftime("%Y%m%d_%H%M%S")])
    else:
        raise ValueError("Unknown format")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
calculate_epsilon.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy import constants

from ..pyrf import resample, ts_scalar


def calculate_epsilon(vdf=None, model_vdf=None, n=None, sc_pot=None, **kwargs):
    """Calculates epsilon parameter using model distribution.

    Parameters
    ----------
    vdf : xarray.Dataset
        Observed particle distribution (skymap).

    model_vdf : xarray.Dataset
        Model particle distribution (skymap).

    n : xarray.DataArray
        Time series of the number density.

    sc_pot : xarray.DataArray
        Time series of the spacecraft potential.

    **kwargs : dict
        Hash table of keyword arguments with :
             * en_channels : list or numpy.ndarray
                Set energy channels to integrate over [min max]; min and max between must be
                between 1 and 32.

    Returns
    -------
    epsilon : xarray.DataArray
        Time series of the epsilon parameter.

    Examples
    --------
    >>> from pyrfu import mms
    >>> eps = mms.calculate_epsilon(vdf, model_vdf, n, sc_pot, en_channel=[4, 32])

    """

    assert vdf is not None and isinstance(vdf, xr.Dataset)
    assert model_vdf is not None and isinstance(model_vdf, xr.Dataset)
    assert n is not None and isinstance(n, xr.DataArray)
    assert sc_pot is not None and isinstance(sc_pot, xr.DataArray)

    if np.abs(np.median(np.diff(vdf.time-n.time))).astype(float) > 0:
        raise RuntimeError("vdf and moments have different times.")

    # Default energy channels used to compute epsilon, lowest energy channel should not be used.
    int_energies = np.arange(2, 33)

    if kwargs.get("en_channels") and isinstance(kwargs["en_channels"], (list, np.ndarray)):
        int_energies = np.arange(kwargs["en_channels"][0], kwargs["en_channels"][1] + 1)

    # Resample sc_pot
    sc_pot = resample(sc_pot, n)

    # Remove zero count points from final calculation
    # model_vdf.data.data[vdf.data.data <= 0] = 0

    model_vdf /= 1e18
    vdf /= 1e18

    vdf_diff = np.abs(vdf.data.data - model_vdf.data.data)

    # Define constants
    qe = constants.e.value

    # Check whether particles are electrons or ions
    if vdf.attrs["specie"] == "e":
        mp = constants.m_e.value
        print("notice : Particles are electrons")
    elif vdf.attrs["specie"] == "i":
        sc_pot.data *= -1
        mp = constants.m_p.value
        print("notice : Particles are electrons")
    else:
        raise ValueError("Invalid specie")

    # Define lengths of variables
    n_ph = len(vdf.phi.data[0, :])

    # Define corrected energy levels using spacecraft potential
    energy_arr = vdf.energy.data
    v, delta_v = [np.zeros(energy_arr.shape) for _ in range(2)]

    for it in range(len(vdf.time)):
        energy_vec, energy_log = [energy_arr[it, :], np.log10(energy_arr[it, :])]

        v[it, :] = np.real(np.sqrt(2 * (energy_vec - sc_pot.data[it]) * qe / mp))

        temp0, temp33 = [2 * energy_log[0] - energy_log[1], 2 * energy_log[-1] - energy_log[-2]]

        energy_all = np.hstack([temp0, energy_log, temp33])

        diff_en_all = np.diff(energy_all)

        energy_upper = 10 ** (energy_log + diff_en_all[1:34] / 2)
        energy_lower = 10 ** (energy_log - diff_en_all[0:33] / 2)

        v_upper = np.sqrt(2 * qe * energy_upper / mp)
        v_lower = np.sqrt(2 * qe * energy_lower / mp)

        v_lower[np.isnan(v_lower)] = 0
        v_upper[np.isnan(v_upper)] = 0

        delta_v[it, :] = (v_upper - v_lower)

    v[v < 0] = 0

    # Calculate density of vdf_diff
    delta_ang = (11.25 * np.pi / 180) ** 2
    theta_k = vdf.theta.data

    epsilon, m_psd2_n = [np.zeros(len(vdf.time)), np.ones(n_ph) * np.sin(theta_k * np.pi / 180)]

    for it in range(len(vdf.time)):
        for ii in int_energies:
            tmp = np.squeeze(vdf_diff[it, ii, ...])

            epsilon[it] += np.nansum(np.nansum(tmp * m_psd2_n, axis=0), axis=0) * v[it, ii] ** 2 * \
                           delta_v[it, ii] * delta_ang

    epsilon /= 1e6 * n.data * 2
    epsilon = ts_scalar(vdf.time.data, epsilon)

    return epsilon



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
mms_pl_config.py

@author : Louis RICHARD
"""

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from astropy import constants
from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()

plt.style.use("seaborn-whitegrid")
date_form = mdates.DateFormatter("%H:%M:%S")
sns.set_context("paper")

plt.rc("lines", linewidth=1)
color = np.array([[0, 0, 0], [213, 94, 0], [0, 158, 115], [86, 180, 233]]) / 255


def mms_pl_config(r_mms):
    """Plot spacecraft configuaration with three 2d plots of the position in re and one 3d plot of
    the relative position of the spacecraft:

    Parameters
    ----------
    r_mms : list of xarray.DataArray
        Time series of the spacecraft position

    Returns
    -------
    fig : matplotlib.pyplot.figure
        to fill.

    axs : matplotlib.pyplot.subplotsaxes
        to fill.

    """

    r_earth = constants.r_earth.value / 1000
    r1, r2, r3, r4 = [np.mean(r_xyz, 0) for r_xyz in r_mms]

    r = np.vstack([r1, r2, r3, r4])
    r = np.mean(r, 0)
    dr = r - np.tile(r, (4, 1))

    fig = plt.figure(figsize=(9, 9))
    gs0 = fig.add_gridspec(3, 3, hspace=0.3, left=0.1, right=0.9, bottom=0.1, top=0.9)

    gs00 = gs0[0, :].subgridspec(1, 3, wspace=0.35)
    gs10 = gs0[1:, :].subgridspec(1, 1, wspace=0.35)

    axs0 = fig.add_subplot(gs00[0])
    axs1 = fig.add_subplot(gs00[1])
    axs2 = fig.add_subplot(gs00[2])
    axs3 = fig.add_subplot(gs10[0], projection='3d')

    axs0.scatter(r[0, 0] / r_earth, r[0, 2] / r_earth, marker="s")
    axs0.scatter(r[1, 0] / r_earth, r[1, 2] / r_earth, marker="d")
    axs0.scatter(r[2, 0] / r_earth, r[2, 2] / r_earth, marker="o")
    axs0.scatter(r[3, 0] / r_earth, r[3, 2] / r_earth, marker="^")
    earth = plt.Circle((0, 0), 1, color='k', clip_on=False)
    axs0.add_artist(earth)
    axs0.set_xlim([20, -20])
    axs0.set_ylim([-20, 20])
    axs0.set_aspect("equal")
    axs0.set_xlabel("$X$ [$r_earth$]")
    axs0.set_ylabel("$Z$ [$r_earth$]")
    axs0.set_title("X = {:2.1f} $r_earth$".format(r[0] / r_earth))

    axs1.scatter(r[0, 1] / r_earth, r[0, 2] / r_earth, marker="s")
    axs1.scatter(r[1, 1] / r_earth, r[1, 2] / r_earth, marker="d")
    axs1.scatter(r[2, 1] / r_earth, r[2, 2] / r_earth, marker="o")
    axs1.scatter(r[3, 1] / r_earth, r[3, 2] / r_earth, marker="^")
    earth = plt.Circle((0, 0), 1, color='k', clip_on=False)
    axs1.add_artist(earth)
    axs1.set_xlim([20, -20])
    axs1.set_ylim([-20, 20])
    axs1.set_aspect("equal")
    axs1.set_xlabel("$Y$ [$r_earth$]")
    axs1.set_ylabel("$Z$ [$r_earth$]")
    axs1.set_title("Y = {:2.1f} $r_earth$".format(r[1] / r_earth))

    axs2.scatter(r[0, 0] / r_earth, r[0, 1] / r_earth, marker="s")
    axs2.scatter(r[1, 0] / r_earth, r[1, 1] / r_earth, marker="d")
    axs2.scatter(r[2, 0] / r_earth, r[2, 1] / r_earth, marker="o")
    axs2.scatter(r[3, 0] / r_earth, r[3, 1] / r_earth, marker="^")
    earth = plt.Circle((0, 0), 1, color='k', clip_on=False)
    axs2.add_artist(earth)
    axs2.set_xlim([20, -20])
    axs2.set_ylim([-20, 20])
    axs2.set_aspect("equal")
    axs2.set_xlabel("$X$ [$r_earth$]")
    axs2.set_ylabel("$Y$ [$r_earth$]")
    axs2.set_title("Z = {:2.1f} $r_earth$".format(r[2] / r_earth))

    axs3.view_init(elev=13, azim=-20)
    axs3.scatter(dr[0, 0], dr[0, 1], dr[0, 2], s=50, marker="s")
    axs3.scatter(dr[1, 0], dr[1, 1], dr[1, 2], s=50, marker="d")
    axs3.scatter(dr[2, 0], dr[2, 1], dr[2, 2], s=50, marker="o")
    axs3.scatter(dr[3, 0], dr[3, 1], dr[3, 2], s=50, marker="^")

    axs3.plot([dr[0, 0]] * 2, [dr[0, 1]] * 2, color=color[0], marker="s", zdir='z', zs=-30)
    axs3.plot([dr[1, 0]] * 2, [dr[1, 1]] * 2, color=color[1], marker="d", zdir='z', zs=-30)
    axs3.plot([dr[2, 0]] * 2, [dr[2, 1]] * 2, color=color[2], marker="o", zdir='z', zs=-30)
    axs3.plot([dr[3, 0]] * 2, [dr[3, 1]] * 2, color=color[3], marker="^", zdir='z', zs=-30)

    axs3.plot([dr[0, 0]] * 2, [dr[0, 2]] * 2, color=color[0], marker="s", zdir='y', zs=-30)
    axs3.plot([dr[1, 0]] * 2, [dr[1, 2]] * 2, color=color[1], marker="d", zdir='y', zs=-30)
    axs3.plot([dr[2, 0]] * 2, [dr[2, 2]] * 2, color=color[2], marker="o", zdir='y', zs=-30)
    axs3.plot([dr[3, 0]] * 2, [dr[3, 2]] * 2, color=color[3], marker="^", zdir='y', zs=-30)

    axs3.plot([dr[0, 1]] * 2, [dr[0, 2]] * 2, color=color[0], marker="s", zdir='x', zs=-30)
    axs3.plot([dr[1, 1]] * 2, [dr[1, 2]] * 2, color=color[1], marker="d", zdir='x', zs=-30)
    axs3.plot([dr[2, 1]] * 2, [dr[2, 2]] * 2, color=color[2], marker="o", zdir='x', zs=-30)
    axs3.plot([dr[3, 1]] * 2, [dr[3, 2]] * 2, color=color[3], marker="^", zdir='x', zs=-30)

    axs3.plot([dr[0, 0]] * 2, [dr[0, 1]] * 2, [-30, dr[0, 2]], 'k--', linewidth=.5)
    axs3.plot([dr[1, 0]] * 2, [dr[1, 1]] * 2, [-30, dr[1, 2]], 'k--', linewidth=.5)
    axs3.plot([dr[2, 0]] * 2, [dr[2, 1]] * 2, [-30, dr[2, 2]], 'k--', linewidth=.5)
    axs3.plot([dr[3, 0]] * 2, [dr[3, 1]] * 2, [-30, dr[3, 2]], 'k--', linewidth=.5)

    axs3.plot([dr[0, 0]] * 2, [-30, dr[0, 1]], [dr[0, 2]] * 2, 'k--', linewidth=.5)
    axs3.plot([dr[1, 0]] * 2, [-30, dr[1, 1]], [dr[1, 2]] * 2, 'k--', linewidth=.5)
    axs3.plot([dr[2, 0]] * 2, [-30, dr[2, 1]], [dr[2, 2]] * 2, 'k--', linewidth=.5)
    axs3.plot([dr[3, 0]] * 2, [-30, dr[3, 1]], [dr[3, 2]] * 2, 'k--', linewidth=.5)

    axs3.plot([-30, dr[0, 0]], [dr[0, 1]] * 2, [dr[0, 2]] * 2, 'k--', linewidth=.5)
    axs3.plot([-30, dr[1, 0]], [dr[1, 1]] * 2, [dr[1, 2]] * 2, 'k--', linewidth=.5)
    axs3.plot([-30, dr[2, 0]], [dr[2, 1]] * 2, [dr[2, 2]] * 2, 'k--', linewidth=.5)
    axs3.plot([-30, dr[3, 0]], [dr[3, 1]] * 2, [dr[3, 2]] * 2, 'k--', linewidth=.5)

    axs3.plot(dr[[0, 1], 0], dr[[0, 1], 1], dr[[0, 1], 2], 'k-')
    axs3.plot(dr[[1, 2], 0], dr[[1, 2], 1], dr[[1, 2], 2], 'k-')
    axs3.plot(dr[[2, 0], 0], dr[[2, 0], 1], dr[[2, 0], 2], 'k-')
    axs3.plot(dr[[0, 3], 0], dr[[0, 3], 1], dr[[0, 3], 2], 'k-')
    axs3.plot(dr[[1, 3], 0], dr[[1, 3], 1], dr[[1, 3], 2], 'k-')
    axs3.plot(dr[[2, 3], 0], dr[[2, 3], 1], dr[[2, 3], 2], 'k-')

    axs3.set_xlim([-30, 30])
    axs3.set_ylim([30, -30])
    axs3.set_zlim([-30, 30])
    axs3.set_xlabel("$\\Delta X$ [km]")
    axs3.set_ylabel("$\\Delta Y$ [km]")
    axs3.set_zlabel("$\\Delta Z$ [km]")

    axs3.legend(["MMS1", "MMS2", "MMS3", "MMS4"], frameon=False)

    axs = [axs0, axs1, axs2, axs3]

    return fig, axs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
dec_par_perp.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .resample import resample
from .dot import dot


def dec_par_perp(inp=None, b0=None, flag_spin_plane=False):
    """Decomposes a vector into par/perp to B components. If flagspinplane decomposes components to
    the projection of ``b0`` into the XY plane. ``alpha`` gives the angle between ``b0`` and the XY.
    plane.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the field to decompose.

    b0 : xarray.DataArray
        Time series of the background magnetic field.

    flag_spin_plane : bool, optional
        Flag if True gives the projection in XY plane.

    Returns
    -------
    a_para : xarray.DataArray
        Time series of the input field parallel to the background magnetic field.

    a_perp : xarray.DataArray
        Time series of the input field perpendicular to the background magnetic field.

    alpha : xarray.DataArray
        Time series of the angle between the background magnetic field and the XY plane.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field (FGM) and electric field (EDP)
    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, mms_id)
    >>> # Decompose e_xyz into parallel and perpendicular to b_xyz components
    >>> e_para, e_perp, _ = pyrf.dec_par_perp(e_xyz, b_xyz)

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert b0 is not None and isinstance(b0, xr.DataArray)

    if not flag_spin_plane:
        b_mag = np.linalg.norm(b0, axis=1, keepdims=True)

        ii = np.where(b_mag < 1e-3)[0]

        if ii.size > 0:
            b_mag[ii] = np.ones(len(ii))*1e-3

        b_hat = b0 / b_mag
        b_hat = resample(b_hat, inp)

        a_para = dot(b_hat, inp)
        a_perp = inp.data - (b_hat * np.tile(a_para.data, (3, 1)).T)
        alpha = []
    else:
        b0 = resample(b0, inp)
        bt = np.sqrt(b0[:, 0] ** 2 + b0[:, 1] ** 2)
        b0 /= bt[:, np.newaxis]

        a_para = inp[:, 0] * b0[:, 0] + inp[:, 1] * b0[:, 1]
        a_perp = inp[:, 0] * b0[:, 1] - inp[:, 1] * b0[:, 0]
        alpha = np.arctan2(b0[:, 2], bt)

    return a_para, a_perp, alpha



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
dft_time_shift.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from ..pyrf import calc_fs, ts_scalar


def dft_time_shift(inp=None, tau=None):
    """Shifts the input signal ``inp`` by ``tau`` seconds using discrete fourier transform (DFT).
    Particularly useful when calculating the frequency-wavenumber spectrum of the mms' spin-plane
    or axial probes.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series to be shifted (Note : Only tensor order 1).

    tau : float
        Applied shift in seconds.

    Returns
    -------
    out : xarray.DataArray
        Time series of the shifted input.

    See also
    --------
    pyrfu.mms.fk_power_spectrum : Calculates the frequency-wave number power spectrum.

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert tau is not None and isinstance(tau, float)

    time, sig = [inp.time.data, inp.data]

    # Sampling frequency
    fs = calc_fs(inp)

    # Applied delay in samples.
    ds = np.floor(tau * fs)

    # Forward FFT
    u = np.fft.fft(sig)
    n_p = len(sig)

    # Disregard Nyquist frequency for even-sized dft
    if not n_p % 2:
        u[int(n_p / 2 + 1)] = 0

    f = ((np.arange(n_p) + np.floor(n_p / 2)) % n_p - np.floor(n_p / 2)) / n_p

    # Backward FFT
    out_data = np.fft.ifft(u * np.exp(-2j * np.pi * ds * f))

    out_time = time + int(tau * 1e9)

    out = ts_scalar(out_time, out_data, attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
cart2sph.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from . import ts_vec_xyz


def cart2sph(inp=None, direction_flag=1):
    """Computes magnitude, theta and phi angle from column vector xyz (first column is x ....)
    theta is 0 at equator. direction_flag = -1  -> to make transformation in opposite direction

    Parameters
    ----------
    inp : xarray.DataArray
        Time series to convert.

    direction_flag : {1, -1}
        Set to 1 (default) to transform from cartesian to spherical coordinates.
        Set to -1 to transform from spherical to cartesian coordinates.

    Returns
    -------
    out : xarray.DataArray
        Input field in spherical/cartesian coordinate system.

    """

    assert inp is not None and isinstance(inp, xr.DataArray)

    if inp.attrs["TENSOR_ORDER"] != 1 or inp.data.ndim != 2:
        raise TypeError("Input must be vector field")

    xyz = inp.data

    if direction_flag == -1:
        r = xyz[:, 0]

        st = np.sin(xyz[:, 1] * np.pi / 180)
        ct = np.cos(xyz[:, 1] * np.pi / 180)
        sp = np.sin(xyz[:, 2] * np.pi / 180)
        cp = np.cos(xyz[:, 2] * np.pi / 180)

        z = r * st
        x = r * ct * cp
        y = r * ct * sp

        out_data = np.hstack([x, y, z])

    else:
        xy = xyz[:, 0] ** 2 + xyz[:, 1] ** 2

        r = np.sqrt(xy + xyz[:, 2] ** 2)
        t = np.arctan2(xyz[:, 2], np.sqrt(xy)) * 180 / np.pi
        p = np.arctan2(xyz[:, 1], xyz[:, 0]) * 180 / np.pi

        out_data = np.hstack([r, t, p])

    out = ts_vec_xyz(inp.time.data, out_data, inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
gradient.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .calc_dt import calc_dt


def gradient(inp=None):
    """Computes time derivative of the input variable.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input variable.

    Returns
    -------
    out : xarray.DataArray
        Time series of the time derivative of the input variable.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2017-07-18T13:03:34.000", "2017-07-18T13:07:00.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field
    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> # Time derivative of the magnetic field
    >>> db_dt = pyrf.gradient(b_xyz)

    """

    assert inp is not None and isinstance(inp, xr.DataArray)

    # guess time step
    dt = calc_dt(inp)

    d_inp_dt = np.gradient(inp.data, axis=0) / dt

    out = xr.DataArray(d_inp_dt, coords=inp.coords, dims=inp.dims, attrs=inp.attrs)

    if "UNITS" in out.attrs:
        out.attrs["UNITS"] += "/s"

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
load_ancillary.py

@author : Louis RICHARD
"""

import os
import re
import json
import glob
import bisect
import fnmatch
import pandas as pd

from dateutil import parser as date_parser

from .mms_config import CONFIG


def load_ancillary(level_and_dtype="", tint=None, probe=1, verbose=True):
    """Load ancillary data

    Parameters
    ----------
    level_and_dtype : str
        Ancillary type :
            * predatt
            * predeph
            * defatt
            * defeph

    tint : list of str
        Time interval

    probe : int
        Spacecraft index

    verbose : bool
        Set to True to follow the loading. Default is True

    Returns
    -------
    out : xarray.Dataset
        Time series of the ancillary data

    """
    assert tint is not None

    data_path = CONFIG["local_data_dir"]

    if isinstance(probe, (int, float)):
        probe = str(probe)

    # directory and file name search patterns
    # For now
    # -all ancillary data is in one directory:
    #       mms\ancillary
    # -assume file names are of the form:
    #   SPACECRAFT_FILETYPE_startDate_endDate.version
    #   where SPACECRAFT is [MMS1, MMS2, MMS3, MMS4] in uppercase
    #   and FILETYPE is either DEFATT, PREDATT, DEFEPH, PREDEPH in uppercase
    #   and start/endDate is YYYYDOY
    #   and version is Vnn (.V00, .V01, etc..)
    dir_pattern = os.sep.join([data_path, "ancillary", "mms{}".format(probe), level_and_dtype])
    file_pattern = "_".join(["MMS{}".format(probe), level_and_dtype.upper(), "???????_???????.V??"])

    files_in_tint = []
    out_files = []

    files = glob.glob(os.sep.join([dir_pattern, file_pattern]))

    # find the files within the time interval
    file_regex = re.compile(os.sep.join([dir_pattern,
                                         'MMS' + probe + '_' + level_and_dtype.upper()
                                         + '_([0-9]{7})_([0-9]{7}).V[0-9]{2}']))
    for file in files:
        time_match = file_regex.match(file)
        if time_match is not None:
            start_time = pd.to_datetime(time_match.group(1), format="%Y%j")
            end_time = pd.to_datetime(time_match.group(2), format="%Y%j")
            if start_time < date_parser.parse(tint[1]) and end_time >= date_parser.parse(tint[0]):
                files_in_tint.append(file)

    # ensure only the latest version of each file is loaded
    for file in files_in_tint:
        this_file = file[0:-3] + "V??"
        versions = fnmatch.filter(files_in_tint, this_file)
        if len(versions) > 1:
            out_files.append(sorted(versions)[-1])  # only grab the latest version
        else:
            out_files.append(versions[0])

    files_names = list(set(out_files))
    files_names.sort()

    # Read length of header and columns names from .json file
    with open("./ancillary.json") as f:
        anc_dict = json.load(f)

    if verbose:
        print("Loading ancillary {} files...".format(level_and_dtype))

    df_dict = {}

    for i, file in enumerate(files_names):
        rows = pd.read_csv(file, delim_whitespace=True, header=None,
                           skiprows=anc_dict[level_and_dtype]["header"])

        # Remove footer
        rows = rows[:][:-1]

        # Convert time
        rows[0] = pd.to_datetime(rows[0], format=anc_dict[level_and_dtype]["time_format"])

        start_idx = bisect.bisect_left(rows[0][:], date_parser.parse(tint[0]))
        end_idx = bisect.bisect_left(rows[0][:], date_parser.parse(tint[1]))
        rows.columns = anc_dict[level_and_dtype]["columns_names"]

        df_dict[i] = rows[:][start_idx:end_idx]

    df = df_dict[0]

    for k in list(df_dict.keys())[1:]:
        df = df.append(df_dict[k])

    df = df.sort_values(by="time").set_index(["time"])

    return df.to_xarray()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_feeps_active_eyes.py

@author : Louis RICHARD
"""

import datetime
import numpy as np

from dateutil import parser


def get_feeps_active_eyes(var=None, tint=None, mms_id=1):
    """This function returns the FEEPS active eyes, based on date/mms_id/species/rate.
    
    Parameters
    ----------
    var : dict
        Hash table containing parameters.

    tint : list of str
        Time range.

    mms_id : str
        mms_id e.g., '4' for MMS4.

    Returns
    -------
    sensors : dict
        Hash table containing 2 keys :
            * out["top"] : maps to the active top eyes.
            * out["bottom"] : maps to the active bottom eyes.

    Notes
    -----
    1) Burst mode should include all sensors (TOP and BOTTOM) :
        * electrons : [1, 2, 3, 4, 5, 9, 10, 11, 12].
        * ions : [6, 7, 8].

    2) SITL should return (TOP only) :
        * electrons : set_intersection([5, 11, 12], active_eyes).
        * ions : None.

    3) From Drew Turner, 9/7/2017, srvy mode :
        * before 16 August 2017 :
            * electrons : [3, 4, 5, 11, 12].
            * ion : [6, 7, 8].

        * after 16 August 2017 :
            * MMS1 :
                * Top Eyes : 3, 5, 6, 7, 8, 9, 10, 12
                * Bot Eyes : 2, 4, 5, 6, 7, 8, 9, 10
            * MMS2 :
                * Top Eyes : 1, 2, 3, 5, 6, 8, 10, 11
                * Bot Eyes : 1, 4, 5, 6, 7, 8, 9, 11
            * MMS3
                * Top Eyes : 3, 5, 6, 7, 8, 9, 10, 12
                * Bot Eyes : 1, 2, 3, 6, 7, 8, 9, 10
            * MMS4 :
                * Top Eyes : 3, 4, 5, 6, 8, 9, 10, 11
                * Bot Eyes : 3, 5, 6, 7, 8, 9, 10, 12

    """

    assert var is not None and isinstance(var, dict)
    assert tint is not None and isinstance(tint, list)
    assert isinstance(tint[0], str) and isinstance(tint[1], str)
    assert isinstance(mms_id, (int, str)) and int(mms_id) in np.arange(1, 5)

    mms_id = int(mms_id)

    sensors = {}

    if var["tmmode"].lower() == "brst" and var["dtype"].lower() == "electron": 
        sensors["top"] = [1, 2, 3, 4, 5, 9, 10, 11, 12]
        sensors["bottom"] = [1, 2, 3, 4, 5, 9, 10, 11, 12]

    if var["tmmode"].lower() == "brst" and var["dtype"].lower() == "ion": 
        sensors["top"] = [6, 7, 8]
        sensors["bottom"] = [6, 7, 8]

    # old eyes, srvy mode, prior to 16 August 2017
    if var["dtype"].lower() == "electron":
        sensors["top"] = [3, 4, 5, 11, 12]
        sensors["bottom"] = [3, 4, 5, 11, 12]
    else:
        sensors["top"] = [6, 7, 8]
        sensors["bottom"] = [6, 7, 8]

    if isinstance(tint[0], str): 
        start_time = parser.parse(tint[0])
    else:
        start_time = tint[0]

    # srvy mode, after 16 August 2017
    if start_time >= datetime.datetime(2017, 8, 16) and var["tmmode"].lower() == "srvy":
        active_table = {"1-electron": {}, "1-ion": {}, "2-electron": {}, "2-ion": {},
                        "3-electron": {}, "3-ion": {}, "4-electron": {}, "4-ion": {}}

        active_table["1-electron"]["top"] = [3, 5, 9, 10, 12]
        active_table["1-electron"]["bottom"] = [2, 4, 5, 9, 10]

        active_table["1-ion"]["top"] = [6, 7, 8]
        active_table["1-ion"]["bottom"] = [6, 7, 8]

        active_table["2-electron"]["top"] = [1, 2, 3, 5, 10, 11]
        active_table["2-electron"]["bottom"] = [1, 4, 5, 9, 11]

        active_table["2-ion"]["top"] = [6, 8]
        active_table["2-ion"]["bottom"] = [6, 7, 8]

        active_table["3-electron"]["top"] = [3, 5, 9, 10, 12]
        active_table["3-electron"]["bottom"] = [1, 2, 3, 9, 10]

        active_table["3-ion"]["top"] = [6, 7, 8]
        active_table["3-ion"]["bottom"] = [6, 7, 8]

        active_table["4-electron"]["top"] = [3, 4, 5, 9, 10, 11]
        active_table["4-electron"]["bottom"] = [3, 5, 9, 10, 12]

        active_table["4-ion"]["top"] = [6, 8]
        active_table["4-ion"]["bottom"] = [6, 7, 8]
       
        sensors = active_table["{:d}-{}".format(mms_id, var["dtype"].lower())]
        
        if var["lev"].lower() == "sitl":
            sensors["top"] = list(set(sensors["top"]) & {5, 11, 12})
            sensors["bottom"] = []
            return {"top": list(set(sensors["top"]) & {5, 11, 12}), "bottom": []}

    if var["lev"].lower() == "sitl":
        sensors["top"] = [5, 11, 12]
        sensors["bottom"] = []
        
    return sensors



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_eis_allt.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from .list_files import list_files
from .db_get_ts import db_get_ts


def get_eis_allt(tar_var="Flux_extof_proton_srvy_l2", tint=None, mms_id=2, verbose=True):
    """Read energy spectrum of the selected specie in the selected energy range for all telescopes.

    Parameters
    ----------
    tar_var : str
        Key of the target variable like {data_unit}_{dtype}_{specie}_{data_rate}_{data_lvl}.

    tint : list of str
        Time interval.

    mms_id : int or float or str
        Index of the spacecraft.

    verbose : bool
        Set to True to follow the loading. Default is True.

    Returns
    -------
    out : xarray.Dataset
        Dataset containing the energy spectrum of the 6 telescopes of the Energy Ion Spectrometer.

    """

    assert isinstance(tar_var, str)
    assert tint is not None and isinstance(tint, list)
    assert isinstance(tint[0], str) and isinstance(tint[1], str)
    assert isinstance(mms_id, (int, str)) and int(mms_id) in np.arange(1, 5)
    assert isinstance(verbose, bool)

    # Convert mms_id to integer
    mms_id = int(mms_id)

    data_unit, data_type, specie, data_rate, data_lvl = tar_var.split("_")

    var = {"mms_id": mms_id, "inst": "epd-eis", "tmmode": data_rate, "lev": data_lvl}

    if data_type == "electronenergy":
        if specie == "electron":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"mms{mms_id:d}_epd_eis_{data_rate}_{data_type}_{specie}"
        else:
            raise ValueError("invalid specie")

    elif data_type == "extof":
        if specie == "proton":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"mms{mms_id:d}_epd_eis_{data_rate}_{data_type}_{specie}"
        elif specie == "oxygen":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"mms{mms_id:d}_epd_eis_{data_rate}_{data_type}_{specie}"
        elif specie == "alpha":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"mms{mms_id:d}_epd_eis_{data_rate}_{data_type}_{specie}"
        else:
            raise ValueError("invalid specie")

    elif data_type == "phxtof":
        if specie == "proton":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"mms{mms_id:d}_epd_eis_{data_rate}_{data_type}_{specie}"
        elif specie == "oxygen":
            var["dtype"], var["specie"] = [data_type, specie]

            pref = f"mms{mms_id:d}_epd_eis_{data_rate}_{data_type}_{specie}"
        else:
            raise ValueError("Invalid specie")
    else:
        raise ValueError("Invalid data type")

    # EIS includes the version of the files in the cdfname need to read it before.
    files = list_files(tint, mms_id, var)

    file_version = int(files[0].split("_")[-1][1])
    var["version"] = file_version

    if data_unit.lower() in ["flux", "counts", "cps"]:
        suf = "P{:d}_{}_t".format(file_version, data_unit.lower())
    else:
        raise ValueError("Invalid data unit")

    # Name of the data containing index of the probe, instrument, data rate, data level and data
    # type if needed
    dset_name = f"mms{var['mms_id']:d}_{var['inst']}_{var['tmmode']}_{var['lev']}_{var['dtype']}"

    # Names of the energy spectra in the CDF (one for each telescope)
    cdfnames = ["{}_{}{:d}".format(pref, suf, t) for t in range(6)]

    outdict = {}
    for i, cdfname in enumerate(cdfnames):
        scope_key = f"t{i:d}"

        if verbose:
            print(f"Loading {cdfname}...")

        outdict[scope_key] = db_get_ts(dset_name, cdfname, tint)

    # Build Dataset
    out = xr.Dataset(outdict, attrs=var)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
c_4_j.py

@author : Louis RICHARD
"""

import xarray as xr

from astropy import constants

from .avg_4sc import avg_4sc
from .c_4_grad import c_4_grad
from .cross import cross


def c_4_j(r_list=None, b_list=None):
    """Calculate current density :math:`\\mathbf{J}` from using 4 spacecraft technique [4]_, the
    divergence of the magnetic field :math:`\\nabla . \\mathbf{B}`, magnetic field at the center
    of mass of the tetrahedron, :math:`\\mathbf{J}\\times\\mathbf{B}` force, part of the
    divergence of stress associated with curvature :math:`\\nabla.\\mathbf{T}_{shear}` and
    gradient of the magnetic pressure :math:`\\nabla P_b`. Where :

    .. math::

        \\mathbf{J} = \\frac{\\nabla \\times \\mathbf{B}}{\\mu_0}

        \\mathbf{J}\\times\\mathbf{B} = \\nabla.\\mathbf{T}_{shear} + \\nabla P_b

        \\nabla.\\mathbf{T}_{shear} = \\frac{(\\mathbf{B}.\\nabla) \\mathbf{B}}{\\mu_0}

        \\nabla P_b = \\nabla \\frac{B^2}{2\\mu_0}

    The divergence of the magnetic field is current density units as it shows the error on the
    estimation of the current density [5]_ .

    Parameters
    ----------
    r_list : list of xarray.DataArray
        Time series of the spacecraft position [km].

    b_list : list of xarray.DataArray
        Time series of the magnetic field [nT].

    Returns
    -------
    j : xarray.DataArray
        Time series of the current density [A.m^{-2}].

    div_b : xarray.DataArray
        Time series of the divergence of the magnetic field [A.m^{-2}].

    b_avg : xarray.DataArray
        Time series of the magnetic field at the center of mass of the tetrahedron,
        sampled at 1st SC time steps [nT].

    jxb : xarray.DataArray
        Time series of the :math:`\\mathbf{J}\\times\\mathbf{B}` force [T.A].

    div_t_shear : xarray.DataArray
        Time series of the part of the divergence of stress associated with curvature units
        [T A/m^2].

    div_pb : xarray.DataArray
        Time series of the gradient of the magnetic pressure.

    See also
    --------
    pyrfu.pyrf.c_4_k : Calculates reciprocal vectors in barycentric coordinates.
    pyrfu.pyrf.c_4_grad : Calculate gradient of physical field using 4 spacecraft technique.

    References
    ----------
    .. [4]	Dunlop, M. W., A. Balogh, K.-H. Glassmeier, and P. Robert (2002a), Four-point Cluster
            application of magnetic field analysis tools: The Curl- ometer, J. Geophys. Res.,
            107(A11), 1384, doi : https://doi.org/10.1029/2001JA005088.

    .. [5]	Robert, P., et al. (1998), Accuracy of current determination, in Analysis Methods for
            Multi-Spacecraft Data, edited by G. Paschmann and P. W. Daly, pp. 395–418, Int. Space S
            ci. Inst., Bern. url : http://www.issibern.ch/forads/sr-001-16.pdf

    Examples
    --------
    >>> import numpy as np
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft indices
    >>> mms_list = np.arange(1,5)
    >>> # Load magnetic field and spacecraft position
    >>> b_mms = [mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id) for mms_id in mms_list]
    >>> r_mms = [mms.get_data("R_gse", tint, mms_id) for mms_id in mms_list]
    >>> j, divB, B, jxB, divTshear, divPb = pyrf.c_4_j(r_mms, b_mms)

    """
    assert r_list is not None and isinstance(r_list, list) and len(r_list) == 4
    assert b_list is not None and isinstance(b_list, list) and len(b_list) == 4

    for i in range(4):
        if not isinstance(r_list[i], xr.DataArray):
            raise TypeError("Spacecraft position must be DataArray")

        if not isinstance(b_list[i], xr.DataArray):
            raise TypeError("Magnetic field must be DataArray")

    mu0 = constants.mu0.value

    b_avg = avg_4sc(b_list)

    # Estimate divB/mu0. unit is A/m2
    div_b = c_4_grad(r_list, b_list, "div")

    # to get right units
    div_b *= 1.0e-3 * 1e-9 / mu0

    # estimate current j [A/m2]
    curl_b = c_4_grad(r_list, b_list, "curl")

    # to get right units [A.m^{-2}]
    j = curl_b * 1.0e-3 * 1e-9 / mu0

    # estimate jxB force [T A/m2]
    jxb = 1e-9 * cross(j, b_avg)

    # estimate divTshear = (1/muo) (B*div)B [T A/m2]
    b_div_b = c_4_grad(r_list, b_list, "bdivb")

    # to get right units [T.A.m^{-2}]
    div_t_shear = b_div_b * 1.0e-3 * 1e-9 * 1e-9 / mu0

    # estimate divPb = (1/muo) grad (B^2/2) = divTshear-jxB
    div_pb = div_t_shear - jxb

    return j, div_b, b_avg, jxb, div_t_shear, div_pb



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
psd_moments.py

@author : Louis RICHARD
"""

import bisect
import multiprocessing as mp
import numpy as np
import xarray as xr

from astropy import constants

from ..pyrf.resample import resample
from ..pyrf.ts_scalar import ts_scalar
from ..pyrf.ts_vec_xyz import ts_vec_xyz
from ..pyrf.ts_tensor_xyz import ts_tensor_xyz


# noinspection PyUnboundLocalVariable
def calc_moms(nt, arguments):
    """

    """
    if len(arguments) > 13:
        [is_brst_data, flag_same_e, flag_de, step_table, energy0, delta_v0, energy1, delta_v1,
         qe, sc_pot, p_mass, flag_inner_electron, w_inner_electron, phi_tr, theta_k, int_energies,
         vdf, delta_ang] = arguments
    else:
        [is_brst_data, flag_same_e, flag_de, energy, delta_v, qe, sc_pot, p_mass,
         flag_inner_electron, w_inner_electron, phi_tr, theta_k, int_energies, vdf,
         delta_ang] = arguments

    if is_brst_data:
        if not flag_same_e or not flag_de:
            energy = energy0
            delta_v = delta_v0

            if step_table[nt]:
                energy = energy1
                delta_v = delta_v1

    v = np.real(np.sqrt(2 * qe * (energy - sc_pot.data[nt]) / p_mass))
    v[energy - sc_pot.data[nt] - flag_inner_electron * w_inner_electron < 0] = 0

    if is_brst_data:
        phi_j = phi_tr[nt, :]
    else:
        phi_j = phi_tr

    phi_j = phi_j[:, np.newaxis]

    n_psd = 0
    v_psd = np.zeros(3)
    p_psd = np.zeros((3, 3))
    h_psd = np.zeros(3)

    psd2_n_mat = np.dot(np.ones(phi_j.shape), np.sin(theta_k * np.pi / 180))
    psd2_v_x_mat = -np.dot(np.cos(phi_j * np.pi / 180),
                           np.sin(theta_k * np.pi / 180) * np.sin(theta_k * np.pi / 180))
    psd2_v_y_mat = -np.dot(np.sin(phi_j * np.pi / 180),
                           np.sin(theta_k * np.pi / 180) * np.sin(theta_k * np.pi / 180))
    psd2_v_z_mat = -np.dot(np.ones(phi_j.shape),
                           np.sin(theta_k * np.pi / 180) * np.cos(theta_k * np.pi / 180))
    psd_mf_xx_mat = np.dot(np.cos(phi_j * np.pi / 180) ** 2, np.sin(theta_k * np.pi / 180) ** 3)
    psd_mf_yy_mat = np.dot(np.sin(phi_j * np.pi / 180) ** 2, np.sin(theta_k * np.pi / 180) ** 3)
    psd_mf_zz_mat = np.dot(np.ones(phi_j.shape),
                           np.sin(theta_k * np.pi / 180) * np.cos(theta_k * np.pi / 180) ** 2)
    psd_mf_xy_mat = np.dot(np.cos(phi_j * np.pi / 180) * np.sin(phi_j * np.pi / 180),
                           np.sin(theta_k * np.pi / 180) ** 3)
    psd_mf_xz_mat = np.dot(np.cos(phi_j * np.pi / 180),
                           np.cos(theta_k * np.pi / 180) * np.sin(theta_k * np.pi / 180) ** 2)
    psd_mf_yz_mat = np.dot(np.sin(phi_j * np.pi / 180),
                           np.cos(theta_k * np.pi / 180) * np.sin(theta_k * np.pi / 180) ** 2)

    for ii in int_energies:
        tmp = np.squeeze(vdf[nt, ii, :, :])
        # n_psd_tmp1 = tmp .* psd2_n_mat * v(ii)^2 * delta_v(ii) * delta_ang;
        # n_psd_e32_phi_theta(nt, ii, :, :) = n_psd_tmp1;
        # n_psd_e32(nt, ii) = n_psd_tmp

        # number density
        n_psd_tmp = delta_v[ii] * delta_ang * np.nansum(np.nansum(tmp * psd2_n_mat, axis=0),
                                                        axis=0) * v[ii] ** 2
        n_psd += n_psd_tmp

        # Bulk velocity
        v_temp_x = delta_v[ii] * delta_ang * np.nansum(np.nansum(tmp * psd2_v_x_mat, axis=0),
                                                       axis=0) * v[ii] ** 3
        v_temp_y = delta_v[ii] * delta_ang * np.nansum(np.nansum(tmp * psd2_v_y_mat, axis=0),
                                                       axis=0) * v[ii] ** 3
        v_temp_z = delta_v[ii] * delta_ang * np.nansum(np.nansum(tmp * psd2_v_z_mat, axis=0),
                                                       axis=0) * v[ii] ** 3

        v_psd[0] += v_temp_x
        v_psd[1] += v_temp_y
        v_psd[2] += v_temp_z

        # Pressure tensor
        p_temp_xx = delta_v[ii] * delta_ang * np.nansum(np.nansum(tmp * psd_mf_xx_mat, axis=0),
                                                        axis=0) * v[ii] ** 4
        p_temp_xy = delta_v[ii] * delta_ang * np.nansum(np.nansum(tmp * psd_mf_xy_mat, axis=0),
                                                        axis=0) * v[ii] ** 4
        p_temp_xz = delta_v[ii] * delta_ang * np.nansum(np.nansum(tmp * psd_mf_xz_mat, axis=0),
                                                        axis=0) * v[ii] ** 4
        p_temp_yy = delta_v[ii] * delta_ang * np.nansum(np.nansum(tmp * psd_mf_yy_mat, axis=0),
                                                        axis=0) * v[ii] ** 4
        p_temp_yz = delta_v[ii] * delta_ang * np.nansum(np.nansum(tmp * psd_mf_yz_mat, axis=0),
                                                        axis=0) * v[ii] ** 4
        p_temp_zz = delta_v[ii] * delta_ang * np.nansum(np.nansum(tmp * psd_mf_zz_mat, axis=0),
                                                        axis=0) * v[ii] ** 4

        p_psd[0, 0] += p_temp_xx
        p_psd[0, 1] += p_temp_xy
        p_psd[0, 2] += p_temp_xz
        p_psd[1, 1] += p_temp_yy
        p_psd[1, 2] += p_temp_yz
        p_psd[2, 2] += p_temp_zz

        h_psd[0] = v_temp_x * v[ii] ** 2
        h_psd[1] = v_temp_y * v[ii] ** 2
        h_psd[2] = v_temp_z * v[ii] ** 2

    return n_psd, v_psd, p_psd, h_psd


# noinspection PyUnboundLocalVariable
def psd_moments(vdf=None, sc_pot=None, **kwargs):
    """Computes moments from the FPI particle phase-space densities.

    Parameters
    ----------
    vdf : xarray.Dataset
        3D skymap velocity distribution.

    sc_pot : xarray.DataArray
        Time series of the spacecraft potential.

    **kwargs : dict
        Hash table of keyword arguments with :
            * energy_range : list or numpy.ndarray
                Set energy range in eV to integrate over [E_min E_max]. Energy range is applied
                to energy0 and the same elements are used for energy1 to ensure that the same
                number of points are integrated over.

            * no_sc_pot : bool
                Set to 1 to set spacecraft potential to zero. Calculates moments
                without correcting for spacecraft potential.

            * en_channels : list or numpy.ndarray
                Set energy channels to integrate over [min max]; min and max between must be
                between 1 and 32.

            * partial_moments : numpy.ndarray or xarray.DataArray
                Use a binary array (or DataArray) (pmomsarr) to select which psd points are used
                in the moments calculation. pmomsarr must be a binary array (1s and 0s,
                1s correspond to points used). Array (or data of Dataarray) must be the same size
                as vdf.data.

            * inner_electron : {"on", "off"}
                inner_electrontron potential for electron moments.

    Returns
    --------
    n_psd : xarray.DataArray
        Time series of the number density (1rst moment).

    v_psd : xarray.DataArray
        Time series of the bulk velocity (2nd moment).

    p_psd : xarray.DataArray
        Time series of the pressure tensor (3rd moment).

    p2_psd : xarray.DataArray
        Time series of the pressure tensor.

    t_psd : xarray.DataArray
        Time series of the temperature tensor.

    h_psd : xarray.DataArray
        to fill.

    Examples
    --------
    >>> from pyrfu import mms
    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]
    >>> vdf_e = mms.get_data("PDe_fpi_brst_l2", tint, 1)
    >>> sc_pot = mms.get_data("V_edp_brst_l2", tint, 1)
    >>> particle_moments = mms.psd_moments(vdf_e, sc_pot, energy_range=[1, 1000])

    """

    flag_de = False
    flag_same_e = False
    flag_inner_electron = False

    # [eV] sc_pot + w_inner_electron for electron moments calculation; 2018-01-26, wy;
    w_inner_electron = 3.5

    if vdf is None or sc_pot is None:
        raise ValueError("psd_moments requires at least 2 arguments")

    if not isinstance(vdf, xr.Dataset):
        raise TypeError("vdf must be a Dataset")
    else:
        vdf.data.data *= 1e12

    if not isinstance(sc_pot, xr.DataArray):
        raise TypeError("Spacecraft potential must a DataArray")

    # Check if data is fast or burst resolution
    field_name = vdf.attrs["FIELDNAM"]

    if "brst" in field_name:
        is_brst_data = True
        print("notice : Burst resolution data is used")
    elif "brst" in field_name:
        is_brst_data = False
        print("notice : Fast resolution data is used")
    else:
        raise TypeError("Could not identify if data is fast or burst.")

    phi = vdf.phi.data
    theta_k = vdf.theta
    particle_type = vdf.attrs["species"]

    if is_brst_data:
        step_table = vdf.attrs["estep_table"]
        energy0 = vdf.attrs["energy0"]
        energy1 = vdf.attrs["energy1"]
        e_tmp = energy1 - energy0

        if all(e_tmp) == 0:
            flag_same_e = 1
    else:
        energy = vdf.energy
        e_tmp = energy[0, :] - energy[-1, :]

        if all(e_tmp) == 0:
            energy = energy[0, :]
        else:
            raise TypeError("Could not identify if data is fast or burst.")

    # resample sc_pot to same resolution as particle distributions
    sc_pot = resample(sc_pot, vdf.time)

    int_energies = np.arange(32)

    if "energy_range" in kwargs:
        if isinstance(kwargs["energy_range"], (list, np.ndarray)) and \
                len(kwargs["energy_range"]) == 2:
            if not is_brst_data:
                energy0 = energy

            e_min_max = kwargs["energy_range"]
            start_e = bisect.bisect_left(energy0, e_min_max[0])
            stop_e = bisect.bisect_left(energy0, e_min_max[1])

            int_energies = np.arange(start_e, stop_e)
            print("notice : Using partial energy range")

    if "no_sc_pot" in kwargs:
        if isinstance(kwargs["no_sc_pot"], bool) and not kwargs["no_sc_pot"]:
            sc_pot.data = np.zeros(sc_pot.shape)
            print("notice : Setting spacecraft potential to zero")

    if "en_channels" in kwargs:
        if isinstance(kwargs["en_channels"], (list, np.ndarray)):
            int_energies = np.arange(kwargs["en_channels"][0], kwargs["en_channels"][1])

    if "partial_moments" in kwargs:
        partial_moments = kwargs["partial_moments"]
        if isinstance(partial_moments, xr.DataArray):
            partial_moments = partial_moments.data

        # Check size of partial_moments
        if partial_moments.shape == vdf.data.shape:
            sum_ones = np.sum(np.sum(np.sum(np.sum(partial_moments, axis=-1), axis=-1), axis=-1),
                              axis=-1)
            sum_zeros = np.sum(
                np.sum(np.sum(np.sum(-partial_moments + 1, axis=-1), axis=-1), axis=-1), axis=-1)

            if (sum_ones + sum_zeros) == vdf.data.size:
                print("notice : partial_moments is correct. Partial moments will be calculated")
                vdf.data = vdf.data * partial_moments
            else:
                print("notice : All values are not ones and zeros in partial_moments. Full " +
                      "moments will be calculated")
        else:
            print("notice : Size of partial_moments is wrong. Full moments will be calculated")

    if "inner_electron" in kwargs:
        inner_electron_tmp = kwargs["inner_electron"]
        if inner_electron_tmp == "on" and particle_type[0] == "e":
            flag_inner_electron = True

    # Define constants
    qe = constants.e.value
    kb = constants.k_B.value

    if particle_type[0] == "e":
        p_mass = constants.m_e.value
        print("notice : Particles are electrons")
    elif particle_type[0] == "i":
        p_mass = constants.m_p.value
        sc_pot.data = -sc_pot.data
        print("notice : Particles are ions")
    else:
        raise ValueError("Could not identify the particle type")

    p2_psd = np.zeros((len(vdf.time), 3, 3))

    # angle between theta and phi points is 360/32 = 11.25 degrees
    delta_ang = (11.25 * np.pi / 180) ** 2

    if is_brst_data:
        phi_tr = vdf.phi
    else:
        phi_tr = phi
        phi_size = phi_tr.shape

        if phi_size[1] > phi_size[0]:
            phi_tr = phi_tr.T

    if "delta_energy_minus" in vdf.attrs and "delta_energy_plus" in vdf.attrs:
        energy_minus, energy_plus = [vdf.attrs["delta_energy_plus"], vdf.attrs["delta_energy_plus"]]

        flag_de = True
    else:
        energy_minus, energy_plus = [None, None]

    # Calculate speed widths associated with each energy channel.
    if is_brst_data:  # Burst mode energy/speed widths
        if flag_same_e and flag_de:
            energy = energy0
            energy_upper = energy + energy_plus
            energy_lower = energy - energy_minus
            v_upper = np.sqrt(2 * qe * energy_upper / p_mass)
            v_lower = np.sqrt(2 * qe * energy_lower / p_mass)
            delta_v = v_upper - v_lower
        else:
            energy_all = np.hstack([energy0, energy1])
            energy_all = np.log10(np.sort(energy_all))

            if np.abs(energy_all[1] - energy_all[0]) > 1e-4:
                temp0 = 2 * energy_all[0] - energy_all[1]
            else:
                temp0 = 2 * energy_all[1] - energy_all[2]

            if np.abs(energy_all[63] - energy_all[62]) > 1e-4:
                temp65 = 2 * energy_all[63] - energy_all[62]
            else:
                temp65 = 2 * energy_all[63] - energy_all[61]

            energy_all = np.hstack([temp0, energy_all, temp65])
            diff_en_all = np.diff(energy_all)
            energy0upper = 10 ** (np.log10(energy0) + diff_en_all[1:64:2] / 2)
            energy0lower = 10 ** (np.log10(energy0) - diff_en_all[0:63:2] / 2)
            energy1upper = 10 ** (np.log10(energy1) + diff_en_all[2:65:2] / 2)
            energy1lower = 10 ** (np.log10(energy1) - diff_en_all[1:64:2] / 2)

            v0upper = np.sqrt(2 * qe * energy0upper / p_mass)
            v0lower = np.sqrt(2 * qe * energy0lower / p_mass)
            v1upper = np.sqrt(2 * qe * energy1upper / p_mass)
            v1lower = np.sqrt(2 * qe * energy1lower / p_mass)
            delta_v0 = (v0upper - v0lower) * 2.0
            delta_v1 = (v1upper - v1lower) * 2.0

            # delta_v0(1) = delta_v0(1)*2.7
            # delta_v1(1) = delta_v1(1)*2.7

    else:  # Fast mode energy/speed widths
        energy_all = np.log10(energy)
        temp0 = 2 * energy_all[0] - energy_all[1]
        temp33 = 2 * energy_all[31] - energy_all[30]
        energy_all = np.hstack([temp0, energy_all, temp33])
        diff_en_all = np.diff(energy_all)
        energy_upper = 10 ** (np.log10(energy) + diff_en_all[1:33] / 4)
        energy_lower = 10 ** (np.log10(energy) - diff_en_all[0:32] / 4)
        v_upper = np.sqrt(2 * qe * energy_upper / p_mass)
        v_lower = np.sqrt(2 * qe * energy_lower / p_mass)
        delta_v = (v_upper - v_lower) * 2.0
        delta_v[0] = delta_v[0] * 2.7

    theta_k = theta_k.data[np.newaxis, :]

    # New version parallel
    # args brst :
    # (is_brst_data, flag_same_e, flag_de, step_table, energy0, delta_v0, energy1, delta_v1, qe,
    # sc_pot.data, p_mass, flag_inner_electron, w_inner_electron, phi_tr.data, theta_k,
    # int_energies, vdf.data.data, delta_ang)
    #
    # args fast :
    # (is_brst_data, flag_same_e, flag_de, energy, delta_v, qe, sc_pot.data, p_mass,
    # flag_inner_electron, w_inner_electron, phi_tr.data, theta_k, int_energies, vdf.data,
    # delta_ang)

    if is_brst_data:
        arguments = (is_brst_data, flag_same_e, flag_de, step_table, energy0, delta_v0, energy1,
                     delta_v1, qe, sc_pot.data, p_mass, flag_inner_electron, w_inner_electron,
                     phi_tr.data, theta_k, int_energies, vdf.data.data, delta_ang)
    else:
        arguments = (is_brst_data, flag_same_e, flag_de, energy, delta_v, qe, sc_pot.data,
                     p_mass, flag_inner_electron, w_inner_electron, phi_tr.data, theta_k,
                     int_energies, vdf.data, delta_ang)

    pool = mp.Pool(mp.cpu_count())
    res = pool.starmap(calc_moms, [(nt, arguments) for nt in range(len(vdf.time))])
    out = np.vstack(res)

    n_psd = np.array(out[:, 0], dtype="float")
    v_psd = np.vstack(out[:, 1][:])
    p_psd = np.vstack(out[:, 2][:])
    p_psd = np.reshape(p_psd, (len(n_psd), 3, 3))
    h_psd = np.vstack(out[:, 3][:])

    pool.close()

    # Compute moments in SI units
    p_psd *= p_mass
    v_psd /= n_psd[:, np.newaxis]
    p2_psd[:, 0, 0] = p_psd[:, 0, 0]
    p2_psd[:, 0, 1] = p_psd[:, 0, 1]
    p2_psd[:, 0, 2] = p_psd[:, 0, 2]
    p2_psd[:, 1, 1] = p_psd[:, 1, 1]
    p2_psd[:, 1, 2] = p_psd[:, 1, 2]
    p2_psd[:, 2, 2] = p_psd[:, 2, 2]
    p2_psd[:, 1, 0] = p2_psd[:, 0, 1]
    p2_psd[:, 2, 0] = p2_psd[:, 0, 2]
    p2_psd[:, 2, 1] = p2_psd[:, 1, 2]

    p_psd[:, 0, 0] -= p_mass * n_psd * v_psd[:, 0] * v_psd[:, 0]
    p_psd[:, 0, 1] -= p_mass * n_psd * v_psd[:, 0] * v_psd[:, 1]
    p_psd[:, 0, 2] -= p_mass * n_psd * v_psd[:, 0] * v_psd[:, 2]
    p_psd[:, 1, 1] -= p_mass * n_psd * v_psd[:, 1] * v_psd[:, 1]
    p_psd[:, 1, 2] -= p_mass * n_psd * v_psd[:, 1] * v_psd[:, 2]
    p_psd[:, 2, 2] -= p_mass * n_psd * v_psd[:, 2] * v_psd[:, 2]
    p_psd[:, 1, 0] = p_psd[:, 0, 1]
    p_psd[:, 2, 0] = p_psd[:, 0, 2]
    p_psd[:, 2, 1] = p_psd[:, 1, 2]

    p_trace = np.trace(p_psd, axis1=1, axis2=2)
    t_psd = np.zeros(p_psd.shape)
    t_psd[...] = p_psd[...] / (kb * n_psd[:, np.newaxis, np.newaxis])
    t_psd[:, 1, 0] = t_psd[:, 1, 0]
    t_psd[:, 2, 0] = t_psd[:, 2, 0]
    t_psd[:, 2, 1] = t_psd[:, 2, 1]

    v_abs2 = np.linalg.norm(v_psd, axis=1) ** 2
    h_psd *= p_mass / 2
    h_psd[:, 0] -= v_psd[:, 0] * p_psd[:, 0, 0] + v_psd[:, 1] * p_psd[:, 0, 1] + v_psd[:,
                                                                                 2] * p_psd[:, 0, 2]
    h_psd[:, 0] -= 0.5 * v_psd[:, 0] * p_trace + 0.5 * p_mass * n_psd * v_abs2 * v_psd[:, 0]
    h_psd[:, 1] -= v_psd[:, 0] * p_psd[:, 0, 1] + v_psd[:, 1] * p_psd[:, 1, 1] + v_psd[:,
                                                                                 2] * p_psd[:, 1, 2]
    h_psd[:, 1] -= 0.5 * v_psd[:, 1] * p_trace + 0.5 * p_mass * n_psd * v_abs2 * v_psd[:, 1]
    h_psd[:, 2] -= v_psd[:, 0] * p_psd[:, 0, 2] + v_psd[:, 1] * p_psd[:, 1, 2] + v_psd[:,
                                                                                 2] * p_psd[:, 2, 2]
    h_psd[:, 2] -= 0.5 * v_psd[:, 2] * p_trace + 0.5 * p_mass * n_psd * v_abs2 * v_psd[:, 2]

    # Convert to typical units (/cc, km/s, nP, eV, and ergs/s/cm^2).
    n_psd /= 1e6
    # n_psd_e32              /= 1e6
    # n_psd_e32_phi_theta    /= 1e6
    v_psd /= 1e3
    p_psd *= 1e9
    p2_psd *= 1e9
    t_psd *= kb / qe
    h_psd *= 1e3

    # Construct TSeries
    n_psd = ts_scalar(vdf.time.data, n_psd)
    # n_psd_e32 = ts_scalar(vdf.time, n_psd_e32);
    # n_psd_skymap = ts_skymap(vdf.time.data, n_psd_e32_phi_theta,energy, phi.data, theta_k);
    v_psd = ts_vec_xyz(vdf.time.data, v_psd)
    p_psd = ts_tensor_xyz(vdf.time.data, p_psd)
    p2_psd = ts_tensor_xyz(vdf.time.data, p2_psd)
    t_psd = ts_tensor_xyz(vdf.time.data, t_psd)
    h_psd = ts_vec_xyz(vdf.time.data, h_psd)

    return n_psd, v_psd, p_psd, p2_psd, t_psd, h_psd



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
end.py

@author : Louis RICHARD
"""

import xarray as xr
from astropy.time import Time


def end(inp=None):
    """Gives the last time of the time series in unix format.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input variable.

    Returns
    -------
    out : float or str
        Value of the last time in the desired format.

    """

    assert inp is not None and isinstance(inp, xr.DataArray)

    out = Time(inp.time.data[-1], format="datetime64").unix

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
median_bins.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def median_bins(x=None, y=None, bins=10):
    """Computes median of values of y corresponding to bins of x

    Parameters
    ----------
    x : xarray.DataArray
        Time series of the quantity of bins.

    y : xarray.DataArray
        Time series of the quantity to the median.

    bins : int
        Number of bins.

    Returns
    -------
    out : xarray.Dataset
        Dataset with :
            * bins : xarray.DataArray
                bin values of the x variable.

            * data : xarray.DataArray
                Median values of y corresponding to each bin of x.

            * sigma : xarray.DataArray
                Standard deviation.

    Examples
    --------
    >>> import numpy
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Spacecraft indices
    >>> mms_list = numpy.arange(1,5)
    >>> # Load magnetic field and electric field
    >>> r_mms, b_mms = [[] * 4 for _ in range(2)]
    >>> for mms_id in range(1, 5):
    >>> 	r_mms.append(mms.get_data("R_gse", tint, mms_id))
    >>> 	b_mms.append(mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id))
    >>>
    >>> # Compute current density, etc
    >>> j_xyz, div_b, b_xyz, jxb, div_t_shear, div_pb = pyrf.c_4_j(r_mms, b_mms)
    >>> # Compute magnitude of B and J
    >>> b_mag = pyrf.norm(b_xyz)
    >>> j_mag = pyrf.norm(j_xyz)
    >>> # Median value of |J| for 10 bins of |B|
    >>> med_b_j = pyrf.mean_bins(b_mag, j_mag)

    """

    assert x is not None and isinstance(x, xr.DataArray)
    assert y is None or isinstance(y, xr.DataArray)

    x_sort = np.sort(x.data)
    x_edge = np.linspace(x_sort[0], x_sort[-1], bins + 1)

    m, s = [np.zeros(bins), np.zeros(bins)]

    for i in range(bins):
        idx_l = x.data > x_edge[i]
        idx_r = x.data < x_edge[i+1]

        y_bins = np.abs(y.data[idx_l * idx_r])

        m[i], s[i] = [np.median(y_bins), np.std(y_bins)]

    bins = x_edge[:-1] + np.median(np.diff(x_edge)) / 2

    out_dict = {"data": (["bins"], m), "sigma": (["bins"], s), "bins": bins}

    out = xr.Dataset(out_dict)

    return out
